from __future__ import annotations

from abc import ABC, abstractmethod
import csv
import importlib
import importlib.metadata
import io
from pathlib import Path
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Mapping,
    NamedTuple,
    SupportsIndex,
    overload,
    Sequence,
)
import numpy as np
from himena.consts import ExcelFileTypes
from himena.utils.misc import lru_cache
from himena.types import WidgetDataModel

if TYPE_CHECKING:
    from typing import TypeGuard, Self
    from numpy.typing import NDArray
    import pandas as pd
    import polars as pl
    import pyarrow as pa
    import narwhals as nw


@lru_cache(maxsize=1)
def list_installed_dataframe_packages() -> list[str]:
    """Return a list of installed dataframe package names."""
    installed: list[str] = ["dict"]
    for package_name in ["pandas", "polars", "pyarrow"]:
        if next(importlib.metadata.distributions(name=package_name), None):
            installed.append(package_name)
    return installed


def _read_csv_dict(file) -> dict[str, np.ndarray]:
    if isinstance(file, (str, Path)):
        with open(file) as f:
            return _read_csv_dict(f)
    csv_reader = csv.reader(file)
    header = next(csv_reader)
    data = {k: [] for k in header}
    for row in csv_reader:
        for k, v in zip(header, row):
            data[k].append(v)
    return {k: _as_array(v) for k, v in data.items()}


def _as_array(ar: list[str]) -> np.ndarray:
    try:
        return np.array(ar, dtype=int)
    except Exception:
        ar_str = np.array(ar, dtype=np.dtypes.StringDType())
        try:
            return np.array(np.where(ar_str == "", "nan", ar_str), dtype=float)
        except Exception:
            return ar_str


def read_csv(mod: str, file) -> Any:
    if mod == "dict":
        return _read_csv_dict(file)
    if mod == "pandas":
        return importlib.import_module(mod).read_csv(file, header=0)
    elif mod == "polars":
        return importlib.import_module(mod).read_csv(file, has_header=True)
    elif mod == "pyarrow":
        if isinstance(file, io.StringIO):
            # pyarrow does not support StringIO
            file = io.BytesIO(file.getvalue().encode())
        return importlib.import_module(mod + ".csv").read_csv(file)
    else:
        raise ValueError(f"Unsupported module: {mod}")


def _see_imported_module(arr: Any, module: str, class_name: str = "DataFrame") -> bool:
    typ = type(arr)
    if (
        typ.__name__ != class_name
        or module not in sys.modules
        or typ.__module__.split(".")[0] != module
    ):
        return False
    return True


def is_pandas_dataframe(df) -> TypeGuard[pd.DataFrame]:
    if _see_imported_module(df, "pandas"):
        import pandas as pd

        return isinstance(df, pd.DataFrame)
    return False


def is_polars_dataframe(df) -> TypeGuard[pl.DataFrame]:
    if _see_imported_module(df, "polars"):
        import polars as pl

        return isinstance(df, pl.DataFrame)
    return False


def is_pyarrow_table(df) -> TypeGuard[pa.Table]:
    if _see_imported_module(df, "pyarrow", "Table"):
        import pyarrow as pa

        return isinstance(df, pa.Table)
    return False


def is_narwhals_dataframe(df) -> TypeGuard[nw.DataFrame]:
    if _see_imported_module(df, "narwhals"):
        import narwhals as nw

        return isinstance(df, nw.DataFrame)
    return False


class DataFrameWrapper(ABC):
    def __init__(self, df):
        self._df = df

    def unwrap(self):
        return self._df

    def __repr__(self) -> str:
        return f"{self.type_name()} {self.shape} of data:\n{self._df!r}"

    @overload
    def __getitem__(self, key: tuple[SupportsIndex, SupportsIndex]) -> Any: ...
    @overload
    def __getitem__(self, key: tuple[slice, SupportsIndex]) -> np.ndarray: ...
    @overload
    def __getitem__(self, key: str) -> np.ndarray: ...

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.column_to_array(key)
        elif isinstance(key, tuple):
            r, c = key
            if hasattr(c, "__index__"):
                cindex = int(c)
            else:
                raise TypeError(f"{type(c)} cannot be used for the column index")
            if hasattr(r, "__index__"):
                return self.get_item((r, cindex))
            elif isinstance(r, slice):
                cname = self.column_names()[cindex]
                return self.column_to_array(cname)[r]
            else:
                raise TypeError(f"Cannot slice array with {type(r)}")
        else:
            raise TypeError(f"Unsupported key type: {type(key)}")

    @abstractmethod
    def get_item(self, key: tuple[int, int]) -> Any:
        """Return the value at the given row and column indices"""

    @abstractmethod
    def get_subset(self, r0, r1, c0, c1) -> DataFrameWrapper:
        """Return a subset of the dataframe by slicing at df[r0:r1, c0:c1]."""

    @abstractmethod
    def num_rows(self) -> int:
        """Return the number of rows in the dataframe."""

    @abstractmethod
    def num_columns(self) -> int:
        """Return the number of columns in the dataframe."""

    @abstractmethod
    def column_names(self) -> list[str]:
        """Return the names of the columns in the dataframe."""

    @abstractmethod
    def get_dtype(self, index: int) -> DtypeTuple:
        """Return the dtype of the column at the given index."""

    @classmethod
    @abstractmethod
    def from_csv_string(
        self, str_or_buf: str | io.StringIO, separator: str = ","
    ) -> Self:
        """Create a dataframe from a CSV string."""

    @abstractmethod
    def to_csv_string(self, separator: str = ",") -> str:
        """Convert the dataframe to a CSV string."""

    @abstractmethod
    def to_list(self) -> list[list[Any]]:
        """Convert dataframe to a 2D list"""

    @abstractmethod
    def column_to_array(self, name: str) -> np.ndarray:
        """Return a column of the dataframe as an 1D numpy array."""

    @abstractmethod
    def with_columns(self, data: dict[str, np.ndarray]) -> Self:
        """Set the columns of the dataframe using a dictionary of names and 1D numpy arrays."""

    @classmethod
    @abstractmethod
    def from_dict(cls, data: dict[str, np.ndarray]) -> Self:
        """Create a dataframe from a dictionary of column names and arrays."""

    def to_dict(self) -> dict[str, np.ndarray]:
        return {k: self.column_to_array(k) for k in self.column_names()}

    def copy(self) -> Self:
        """Return a shallow copy of the dataframe."""
        return self.from_dict(self.to_dict())

    def type_name(self) -> str:
        mod = type(self._df).__module__.split(".")[0]
        return f"{mod}.{type(self._df).__name__}"

    @property
    def dtypes(self) -> list[DtypeTuple]:
        return [self.get_dtype(i) for i in range(self.num_columns())]

    @property
    def shape(self) -> tuple[int, int]:
        return self.num_rows(), self.num_columns()

    @abstractmethod
    def write(self, file: str | Path):
        """Write the dataframe to a file."""

    def __len__(self) -> int:
        return self.num_rows()

    def filter(self, array: NDArray[np.bool_] | Sequence[int]) -> Self:
        """Filter the dataframe by the given boolean array or indices."""
        dict_filt = {k: v[array] for k, v in self.to_dict().items()}
        return self.from_dict(dict_filt)

    def sort(self, key: str, *, descending: bool = False) -> Self:
        """Sort the dataframe by the given key."""
        d = self.to_dict()
        order = np.argsort(d[key])
        if descending:
            order = len(order) - 1 - order
        dict_sorted = {k: v[order] for k, v in d.items()}
        return self.from_dict(dict_sorted)

    def select(self, columns: list[str]) -> Self:
        """Select columns by name."""
        dict_new = {k: v for k, v in self.to_dict().items() if k in columns}
        df_new = self.from_dict(dict_new)
        return df_new


class DictWrapper(DataFrameWrapper):
    def __init__(self, df: Mapping[str, np.ndarray]):
        self._df = df
        self._columns = list(df.keys())

    def get_item(self, key: tuple[int, int]) -> Any:
        r, c = key
        col_name = self._columns[c]
        return self._df[col_name][r]

    def get_subset(self, r0, r1, c0, c1) -> DictWrapper:
        keys = self._columns[c0:c1]
        return DictWrapper({k: self._df[k][r0:r1] for k in keys})

    def num_rows(self) -> int:
        return len(next(iter(self._df.values()), []))

    def num_columns(self) -> int:
        return len(self._columns)

    def column_names(self) -> list[str]:
        return self._columns

    def get_dtype(self, index: int) -> DtypeTuple:
        col_name = self._columns[index]
        dtype = self._df[col_name].dtype
        return DtypeTuple(str(dtype), dtype.kind)

    @classmethod
    def from_csv_string(self, str_or_buf: str | io.StringIO, separator: str = ","):
        if isinstance(str_or_buf, str):
            buf = io.StringIO(str_or_buf)
        else:
            buf = str_or_buf
        csv_reader = csv.reader(buf, delimiter=separator)
        header = next(csv_reader)
        data = {k: [] for k in header}
        for row in csv_reader:
            for k, v in zip(header, row):
                data[k].append(v)
        return DictWrapper({k: _as_array(v) for k, v in data.items()})

    def to_csv_string(self, separator: str = ",") -> str:
        lines = [separator.join(self.column_names())]
        for i in range(self.num_rows()):
            lines.append(
                separator.join(str(self._df[k][i]) for k in self.column_names())
            )
        return "\n".join(lines)

    def to_list(self) -> list[list[Any]]:
        return [
            [self._df[k][i] for k in self.column_names()]
            for i in range(self.num_rows())
        ]

    def column_to_array(self, name: str) -> np.ndarray:
        return np.asarray(self._df[name])

    def with_columns(self, data: dict[str, np.ndarray]) -> DictWrapper:
        new_df = dict(self._df)
        new_df.update(data)
        return DictWrapper(new_df)

    @classmethod
    def from_dict(cls, data: dict) -> DataFrameWrapper:
        content: dict[str, np.ndarray] = {}
        length = -1
        for k, v in data.items():
            v_arr = np.asarray(v)
            if v_arr.ndim == 1:
                if length < 0:
                    length = len(v_arr)
                elif length != v_arr.size:
                    raise ValueError(
                        "All arrays must have the same length. Consensus length is "
                        f"{length} but got {v_arr.size} for {k!r}."
                    )
            elif v_arr.ndim > 1:
                # this may happen when the series is a list of arrays
                v_arr = np.empty(v_arr.shape[0], dtype=object)
                v_arr[:] = list(v)
            content[k] = v_arr
        if length < 0:  # all arrays are scalar. Interpret as a single-row data frame.
            length = 1
        for k, v in content.items():
            if v.ndim == 0:
                content[k] = np.full(length, v)
        return DictWrapper(content)

    def write(self, file: str | Path):
        path = Path(file)
        if path.suffix in (".csv", ".txt"):
            sep = ","
        elif path.suffix == ".tsv":
            sep = "\t"
        else:
            raise ValueError(f"DictWrapper does not support writing as a {path.suffix}")
        path.write_text(self.to_csv_string(sep))
        return None


class PandasWrapper(DataFrameWrapper):
    def __init__(self, df: pd.DataFrame):
        self._df = df

    def get_item(self, key: tuple[int, int]) -> Any:
        return self._df.iloc[key]

    def get_subset(self, r0, r1, c0, c1) -> PandasWrapper:
        return PandasWrapper(self._df.iloc[r0:r1, c0:c1])

    def num_rows(self) -> int:
        return self._df.shape[0]

    def num_columns(self) -> int:
        return self._df.shape[1]

    def column_names(self) -> list[str]:
        return self._df.columns.tolist()

    def get_dtype(self, index: int) -> DtypeTuple:
        pd_dtype = self._df.dtypes.iloc[index]
        if isinstance(pd_dtype, np.dtype):
            return DtypeTuple(pd_dtype.name, pd_dtype.kind)
        return DtypeTuple(str(pd_dtype), getattr(pd_dtype, "kind", "O"))

    @classmethod
    def from_csv_string(
        cls, str_or_buf: str | io.StringIO, separator: str = ","
    ) -> DataFrameWrapper:
        import pandas as pd

        if isinstance(str_or_buf, str):
            str_or_buf = io.StringIO(str_or_buf)
        return PandasWrapper(pd.read_csv(str_or_buf, sep=separator))

    def to_csv_string(self, separator: str = ",") -> str:
        return self._df.to_csv(sep=separator, index=False)

    def to_list(self) -> list[list[Any]]:
        return self._df.values.tolist()

    def column_to_array(self, name: str) -> np.ndarray:
        return self._df[name].to_numpy()

    def with_columns(self, data: dict[str, np.ndarray]) -> PandasWrapper:
        df_new = self._df.assign(**data)
        return PandasWrapper(df_new)

    @classmethod
    def from_dict(cls, data: dict) -> DataFrameWrapper:
        import pandas as pd

        return PandasWrapper(pd.DataFrame(data))

    def write(self, file: str | Path):
        path = Path(file)
        if path.suffix == ".tsv":
            self._df.to_csv(path, sep="\t")
        elif path.suffix == ".parquet":
            self._df.to_parquet(path)
        elif path.suffix == ".feather":
            self._df.to_feather(path)
        elif path.suffix == ".json":
            self._df.to_json(path)
        elif path.suffix in (".html", ".htm"):
            self._df.to_html(path)
        elif path.suffix in ExcelFileTypes:
            self._df.to_excel(path)
        elif path.suffix == ".pickle":
            self._df.to_pickle(path)
        elif path.suffix == ".md":
            self._df.to_markdown(path)
        elif path.suffix in (".csv", ".txt"):
            self._df.to_csv(path)
        else:
            raise ValueError(
                "Cannot write a pandas dataframe to a file with the given extension "
                f"{path.suffix!r}"
            )


class PolarsWrapper(DataFrameWrapper):
    def __init__(self, df: pl.DataFrame):
        self._df = df

    def get_item(self, key: tuple[int, int]) -> Any:
        return self._df[key]

    def get_subset(self, r0, r1, c0, c1) -> PolarsWrapper:
        return PolarsWrapper(self._df[r0:r1, c0:c1])

    def num_rows(self) -> int:
        return self._df.shape[0]

    def num_columns(self) -> int:
        return self._df.shape[1]

    def column_names(self) -> list[str]:
        return self._df.columns

    def get_dtype(self, index: int) -> DtypeTuple:
        import polars as pl

        pl_dtype = self._df.dtypes[index]
        if pl_dtype == pl.Int8:
            return DtypeTuple("Int8", "i")
        if pl_dtype == pl.Int16:
            return DtypeTuple("Int16", "i")
        if pl_dtype == pl.Int32:
            return DtypeTuple("Int32", "i")
        if pl_dtype == pl.Int64:
            return DtypeTuple("Int64", "i")
        if pl_dtype == pl.UInt8:
            return DtypeTuple("UInt8", "u")
        if pl_dtype == pl.UInt16:
            return DtypeTuple("UInt16", "u")
        if pl_dtype == pl.UInt32:
            return DtypeTuple("UInt32", "u")
        if pl_dtype == pl.UInt64:
            return DtypeTuple("UInt64", "u")
        if pl_dtype == pl.Float32:
            return DtypeTuple("Float32", "f")
        if pl_dtype == pl.Float64:
            return DtypeTuple("Float64", "f")
        if pl_dtype == pl.Boolean:
            return DtypeTuple("Boolean", "b")
        return DtypeTuple(str(pl_dtype), "O")

    @classmethod
    def from_csv_string(
        cls, str_or_buf: str | io.StringIO, separator: str = ","
    ) -> PolarsWrapper:
        import polars as pl

        if isinstance(str_or_buf, str):
            str_or_buf = io.StringIO(str_or_buf)
        return PolarsWrapper(pl.read_csv(str_or_buf, separator=separator))

    def to_csv_string(self, separator: str = ",") -> str:
        return self._df.write_csv(separator=separator)

    def to_list(self) -> list[list[Any]]:
        return [list(row) for row in self._df.iter_rows()]

    def column_to_array(self, name: str) -> np.ndarray:
        return self._df[name].to_numpy()

    def with_columns(self, data: dict[str, np.ndarray]) -> PolarsWrapper:
        df_new = self._df.with_columns(**data)
        return PolarsWrapper(df_new)

    @classmethod
    def from_dict(cls, data: dict) -> DataFrameWrapper:
        import polars as pl

        return PolarsWrapper(pl.DataFrame(data))

    def write(self, file: str | Path):
        path = Path(file)
        if path.suffix == ".tsv":
            self._df.write_csv(path, separator="\t")
        elif path.suffix in (".csv", ".txt"):
            self._df.write_csv(path)
        elif path.suffix == ".parquet":
            self._df.write_parquet(path)
        elif path.suffix == ".json":
            self._df.write_json(path)
        elif path.suffix in ExcelFileTypes:
            self._df.write_excel(path)
        else:
            raise ValueError(
                "Cannot write a pandas dataframe to a file with the given extension "
                f"{path.suffix!r}"
            )


class PyarrowWrapper(DataFrameWrapper):
    def __init__(self, df: pa.Table):
        self._df = df

    def get_item(self, key: tuple[int, int]) -> Any:
        r, c = key
        col_name = self._df.column_names[c]
        return self._df[col_name][r].as_py()

    def get_subset(self, r0, r1, c0, c1) -> PyarrowWrapper:
        df_sub = self._df.slice(r0, r1 - r0).select(self._df.column_names[c0:c1])
        return PyarrowWrapper(df_sub)

    def num_rows(self) -> int:
        return self._df.num_rows

    def num_columns(self) -> int:
        return self._df.num_columns

    def column_names(self) -> list[str]:
        return self._df.column_names

    def get_dtype(self, index: int) -> DtypeTuple:
        import pyarrow as pa

        pa_type = self._df.schema[index].type
        if pa_type == pa.int8():
            return DtypeTuple("int8", "i")
        if pa_type == pa.int16():
            return DtypeTuple("int16", "i")
        if pa_type == pa.int32():
            return DtypeTuple("int32", "i")
        if pa_type == pa.int64():
            return DtypeTuple("int64", "i")
        if pa_type == pa.uint8():
            return DtypeTuple("uint8", "u")
        if pa_type == pa.uint16():
            return DtypeTuple("uint16", "u")
        if pa_type == pa.uint32():
            return DtypeTuple("uint32", "u")
        if pa_type == pa.uint64():
            return DtypeTuple("uint64", "u")
        if pa_type == pa.float32():
            return DtypeTuple("float32", "f")
        if pa_type == pa.float64():
            return DtypeTuple("float64", "f")
        if pa_type == pa.bool_():
            return DtypeTuple("bool", "b")
        return DtypeTuple(str(pa_type), "O")

    @classmethod
    def from_csv_string(cls, str_or_buf: str, separator: str = ",") -> PyarrowWrapper:
        import pyarrow as pa

        if isinstance(str_or_buf, str):
            buf = io.BytesIO(str_or_buf.encode())
        else:
            buf = io.BytesIO(str_or_buf.getvalue().encode())
        return PyarrowWrapper(
            pa.csv.read_csv(buf, parse_options=pa.csv.ParseOptions(delimiter=separator))
        )

    def to_csv_string(self, separator: str = ",") -> str:
        lines = [separator.join(self.column_names())]
        for a in self._df.to_pylist():
            a: dict[str, Any]
            lines.append(separator.join(str(cell) for cell in a.values()))
        return "\n".join(lines)

    def to_list(self) -> list[list[Any]]:
        return [list(a.values()) for a in self._df.to_pylist()]

    def column_to_array(self, name: str) -> np.ndarray:
        return self._df[name].to_numpy()

    def with_columns(self, data: dict[str, np.ndarray]) -> PyarrowWrapper:
        import pyarrow as pa

        new_table = self._df
        for name in data:
            new_table = new_table.append_column(name, pa.array(data[name]))
        return PyarrowWrapper(new_table)

    @classmethod
    def from_dict(cls, data: dict) -> DataFrameWrapper:
        import pyarrow as pa

        return PyarrowWrapper(pa.Table.from_pydict(data))

    def write(self, file: str | Path):
        import pyarrow.csv
        import pyarrow.parquet
        import pyarrow.feather

        path = Path(file)
        if path.suffix == ".tsv":
            pyarrow.csv.write_csv(
                self._df, path, write_options=pyarrow.csv.WriteOptions(delimiter="\t")
            )
        elif path.suffix in (".csv", ".txt"):
            pyarrow.csv.write_csv(self._df, path)
        elif path.suffix == ".parquet":
            pyarrow.parquet.write_table(self._df, path)
        elif path.suffix == ".feather":
            pyarrow.feather.write_feather(self._df, path)
        else:
            raise ValueError(
                "Cannot write a pandas dataframe to a file with the given extension "
                f"{path.suffix!r}"
            )


class DtypeTuple(NamedTuple):
    """Normalized dtype description."""

    name: str  # any string representation of dtype
    kind: str  # must follow numpy's kind character


def wrap_dataframe(df) -> DataFrameWrapper:
    if isinstance(df, Mapping):
        return DictWrapper.from_dict(df)
    if is_pandas_dataframe(df):
        return PandasWrapper(df)
    if is_polars_dataframe(df):
        return PolarsWrapper(df)
    if is_pyarrow_table(df):
        return PyarrowWrapper(df)
    if is_narwhals_dataframe(df):
        return wrap_dataframe(df.to_native())
    if isinstance(df, WidgetDataModel):
        return wrap_dataframe(df.value)
    if isinstance(df, DataFrameWrapper):
        return df
    raise TypeError(f"Unsupported dataframe type: {type(df)}")



from __future__ import annotations
from typing import Any

from pydantic import Field
from himena.standards.plotting.components import BasePlotModel, Face, Edge
from himena.standards.plotting import models as _m


class Scatter3D(_m.Scatter):
    """Plot model for a 3D scatter plot."""

    z: Any = Field(..., description="Z-axis values.")


class Line3D(_m.Line):
    """Plot model for a 3D line plot."""

    z: Any = Field(..., description="Z-axis values.")


class Surface3D(BasePlotModel):
    """Plot model for a 3D surface plot."""

    x: Any = Field(..., description="X-axis values.")
    y: Any = Field(..., description="Y-axis values.")
    z: Any = Field(..., description="Z-axis values.")
    face: Face = Field(default_factory=Face, description="Properties of the faces.")
    edge: Edge = Field(default_factory=Edge, description="Properties of the edges.")

    def plot_option_dict(self) -> dict[str, Any]:
        from himena.qt.magicgui import EdgePropertyEdit, FacePropertyEdit

        return {
            "name": {"widget_type": "LineEdit", "value": self.name},
            "face": {"widget_type": FacePropertyEdit, "value": self.face.model_dump()},
            "edge": {"widget_type": EdgePropertyEdit, "value": self.edge.model_dump()},
        }


class Mesh3D(BasePlotModel):
    """Plot model for a 3D mesh plot."""

    vertices: Any = Field(..., description="Vertices of the mesh.")
    face_indices: Any = Field(..., description="Face indices of the mesh.")
    face: Face = Field(default_factory=Face, description="Properties of the faces.")
    edge: Edge = Field(default_factory=Edge, description="Properties of the edges.")

    def plot_option_dict(self) -> dict[str, Any]:
        from himena.qt.magicgui import EdgePropertyEdit, FacePropertyEdit

        return {
            "name": {"widget_type": "LineEdit", "value": self.name},
            "face": {"widget_type": FacePropertyEdit, "value": self.face.model_dump()},
            "edge": {"widget_type": EdgePropertyEdit, "value": self.edge.model_dump()},
        }



from __future__ import annotations

import importlib
from himena.utils.entries import iter_plugin_info, is_submodule


def install_plugin(module: str):
    """Convenience function to install a plugin during pytest.

    This function is supposed to be used in a fixture so that a himena plugin will be
    installed to the application during the test session.

    ```python
    ## conftest.py
    import pytest
    from himena.testing import install_plugin

    @pytest.fixture(scope="session", autouse=True)
    def init_pytest(request):
        install_plugin("himena-my-plugin-name")
    ```
    """
    found = False
    for info in iter_plugin_info():
        if info.distribution == module:
            importlib.import_module(info.place)
            found = True
        elif is_submodule(info.place, module):
            importlib.import_module(info.place)
            found = True
    if not found:
        raise ValueError(f"Plugin '{module}' not found.")



from himena.plugins import config_field
from dataclasses import dataclass


@dataclass
class FavoriteCommandsConfig:
    """Configuration for the favorite commands widget."""

    commands: list[str] = config_field(
        default_factory=list,
        tooltip="List of favorite commands.",
        widget_type="ListEdit",
        layout="vertical",
        annotation="list[str]",
    )



from himena import MainWindow
from himena_builtins.qt.history._widget import QCommandHistory

def test_command_added(himena_ui: MainWindow):
    history_widget = QCommandHistory(himena_ui)
    assert history_widget._command_list.model().rowCount() == 0
    himena_ui.exec_action("new-tab")
    assert history_widget._command_list.model().rowCount() == 1

    himena_ui.exec_action("builtins:new-text")
    widget = history_widget._command_list.indexWidget(
        history_widget._command_list.model().index(0, 0)
    )
    widget._make_drag()
    widget._enter_event()
    widget.set_button_visible(True)
    widget.set_button_visible(False)
    widget._emit_copy_clicked()
    widget._emit_run_clicked()
    assert history_widget._command_list._find_index_widget(widget).row() == 0
    history_widget.deleteLater()



from __future__ import annotations

from functools import cache
from typing import TYPE_CHECKING
from pydantic import BaseModel, Field
from himena.types import Rect
from himena.utils.misc import iter_subclasses

if TYPE_CHECKING:
    from typing import Self
    import numpy as np
    from numpy.typing import NDArray


class RoiModel(BaseModel):
    """Base class for ROIs (Region of Interest) in images."""

    name: str | None = Field(None, description="Name of the ROI.")

    def model_dump_typed(self) -> dict:
        return {
            "type": _strip_roi_suffix(type(self).__name__.lower()),
            **self.model_dump(),
        }

    @classmethod
    def construct(cls, typ: str, dict_: dict) -> RoiModel:
        """Construct an instance from a dictionary."""
        model_type = pick_roi_model(typ)
        return model_type.model_validate(dict_)

    def to_mask(self, shape: tuple[int, ...]) -> NDArray[np.bool_]:
        raise NotImplementedError


@cache
def pick_roi_model(typ: str) -> type[RoiModel]:
    """Pick an ROI model class from the given type string"""
    for sub in iter_subclasses(RoiModel):
        if _strip_roi_suffix(sub.__name__.lower()) == typ:
            return sub
    raise ValueError(f"Unknown ROI type: {typ!r}")


def _strip_roi_suffix(typ: str) -> str:
    if typ.endswith("roi"):
        typ = typ[:-3]
    return typ


def default_roi_label(nth: int) -> str:
    """Return a default label for the n-th ROI."""
    return f"ROI-{nth}"


class Roi1D(RoiModel):
    def shifted(self, dx: float) -> Self:
        """Return a new 1D ROI translated by the given amount."""
        raise NotImplementedError


class Roi2D(RoiModel):
    def bbox(self) -> Rect[float]:
        """Return the bounding box of the ROI."""
        raise NotImplementedError

    def shifted(self, dx: float, dy: float) -> Self:
        """Return a new 2D ROI translated by the given amount."""
        raise NotImplementedError


class Roi3D(RoiModel):
    def shifted(self, dx: float, dy: float, dz: float) -> Self:
        """Return a new 3D ROI translated by the given amount."""
        raise NotImplementedError



from pathlib import Path
from himena.profile import load_app_profile
from himena.utils.entries import iter_plugin_info, is_submodule, HimenaPluginInfo
from himena.utils.install_plugin import get_install_tool
from configparser import ConfigParser, NoOptionError
import re


def install_and_uninstall(
    install: list[str], uninstall: list[str], profile: str | None
):
    """Process the --install and --uninstall commands."""
    app_profile = load_app_profile(profile or "default")
    plugins_updated = app_profile.plugins.copy()
    infos_installed: list[HimenaPluginInfo] = []
    infos_uninstalled: list[HimenaPluginInfo] = []

    def _install(plugin_info: HimenaPluginInfo) -> None:
        if plugin_info not in infos_installed:
            plugins_updated.append(plugin_info.place)
            infos_installed.append(plugin_info)

    # first, resolve local plugins
    for plugin_name in install:
        if _is_path_like(plugin_name):
            if not Path(plugin_name).exists():
                raise FileNotFoundError(f"Path {plugin_name!r} does not exist.")
            plugin_info_loc = HimenaPluginInfo.from_local_file(plugin_name)
            if plugin_info_loc.place in app_profile.plugins:
                print(f"Plugin {plugin_name!r} is already installed.")
            else:
                plugins_updated.append(plugin_info_loc.place)
                infos_installed.append(plugin_info_loc)
    for plugin_name in uninstall:
        if _is_path_like(plugin_name):
            plugin_info_loc = HimenaPluginInfo.from_local_file(plugin_name)
            if plugin_info_loc.place in app_profile.plugins:
                plugins_updated.remove(plugin_info_loc.place)
                infos_uninstalled.append(plugin_info_loc)
            else:
                print(f"Plugin {plugin_name!r} is not installed.")

    # then, resolve plugins from the distribution
    for info in iter_plugin_info():
        for plugin_name in install:
            if plugin_name in app_profile.plugins:
                print(f"Plugin {plugin_name!r} is already installed.")
            elif is_submodule(info.place, plugin_name):
                _install(info)
            elif info.distribution == plugin_name:
                _install(info)

        for plugin_name in uninstall:
            if is_submodule(info.place, plugin_name) and info.place in plugins_updated:
                plugins_updated.remove(info.place)
                infos_uninstalled.append(info)
            elif info.distribution == plugin_name:
                if info.place in plugins_updated:
                    plugins_updated.remove(info.place)
                    infos_uninstalled.append(info)
                else:
                    print(f"Plugin {plugin_name!r} is not installed.")
    if infos_installed:
        print("Plugins installed:")
        for info in infos_installed:
            print(f"- {info.name} ({info.place}, v{info.version})")
    if infos_uninstalled:
        print("Plugins uninstalled:")
        for info in infos_uninstalled:
            print(f"- {info.name} ({info.place}, v{info.version})")
    elif len(infos_uninstalled) == 0 and len(infos_installed) == 0:
        print("No plugins are installed or uninstalled.")
    new_prof = app_profile.with_plugins(plugins_updated)
    new_prof.save()


def get_and_install(packages: list[str], profile: str | None):
    """Get the packages from PyPI, git, local file etc. and install them."""
    result = get_install_tool().install(packages)
    if result.returncode != 0:
        raise RuntimeError(f"Failed to get: {packages}")
    packages_normed: list[str] = []
    for package in packages:
        if package.endswith(".git") or package.startswith(("http://", "https://")):
            print(
                "Cannot determine the package name from URL yet. Please manually run\n"
                f"  himena {profile or 'default'} --install <package>"
            )
            continue
        elif Path(package).exists():
            if module_name := _extract_module_name(Path(package).resolve()):
                packages_normed.append(module_name)
            else:
                print(
                    f"Cannot determine the package name from the directory {package}. "
                    "Please manually run\n"
                    f"  himena {profile or 'default'} --install <package>"
                )
        else:
            packages_normed.append(package)
    install_and_uninstall(packages_normed, [], profile)


def uninstall_outdated(profile: str | None):
    app_profile = load_app_profile(profile or "default")
    plugins_updated = app_profile.plugins.copy()
    plugins_outdated: list[str] = []

    for plugin_name in plugins_updated:
        for info in iter_plugin_info():
            if _is_path_like(info.place):
                continue
            if is_submodule(info.place, plugin_name):
                break
            elif info.distribution == plugin_name:
                break
        else:
            plugins_outdated.append(plugin_name)

    if plugins_outdated:
        print("Plugins outdated:")
        for plugin_name in plugins_outdated:
            print(f"- {plugin_name}")

        for plugin_name in plugins_outdated:
            plugins_updated.remove(plugin_name)
        new_prof = app_profile.with_plugins(plugins_updated)
        new_prof.save()
    else:
        print("No outdated plugins found.")


def _is_path_like(name: str) -> bool:
    return "/" in name or "\\" in name


def _extract_module_name(directory: Path) -> str | None:
    """Exctract the module name from a python project directory."""
    if (pyproject_toml := directory.joinpath("pyproject.toml")).exists():
        content = pyproject_toml.read_text()
        match = re.search(r'name\s*=\s*["\']([^"\']+)["\']', content)
        if match:
            return match.group(1)
    if (setup_cfg := directory.joinpath("setup.cfg")).exists():
        cfg = ConfigParser()
        cfg.read(setup_cfg)
        try:
            return cfg.get("metadata", "name")
        except NoOptionError:
            return None
    return None



from __future__ import annotations

from qtpy import QtWidgets as QtW, QtCore
import numpy as np


_SUPPORTED_DTYPE_DEPTH = {
    "int": [8, 16, 32, 64],
    "uint": [8, 16, 32, 64],
    "float": [16, 32, 64, 128],
    "complex": [64, 128, 256],
    "bool": [],
}


class QNumericDTypeEdit(QtW.QWidget):
    valueChanged = QtCore.Signal(object)

    def __init__(self, parent=None):
        super().__init__()
        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(1)
        self._combo_type = QtW.QComboBox()
        self._combo_type.addItems(["int", "uint", "float", "complex", "bool"])
        self._combo_type.setCurrentIndex(0)
        self._combo_depth = QtW.QComboBox()
        _fixed = QtW.QSizePolicy.Policy.Fixed
        self._combo_depth.setSizePolicy(_fixed, _fixed)

        layout.addWidget(self._combo_type)
        layout.addWidget(self._combo_depth)

        self._combo_type.currentTextChanged.connect(self._type_changed)
        self._type_changed(self._combo_type.currentText())  # initialize

        # self.setFixedHeight(28)

    def _type_changed(self, text: str):
        self._combo_depth.clear()
        self._combo_depth.addItems([str(d) for d in _SUPPORTED_DTYPE_DEPTH[text]])
        self._combo_depth.setCurrentIndex(0)

    def _cast_dtype(self, dtype) -> np.dtype:
        return np.dtype(dtype)

    def dtype(self) -> np.dtype:
        return self._cast_dtype(
            f"{self._combo_type.currentText()}{self._combo_depth.currentText()}"
        )

    def set_dtype(self, dtype: str):
        _dtype = self._cast_dtype(dtype)
        self._combo_type.setCurrentText(_dtype.kind)
        self._combo_depth.setCurrentText(str(_dtype.itemsize * 8))



from pathlib import Path
from himena import io_utils
from himena.standards import BaseMetadata, write_metadata
from himena.types import WidgetDataModel
from himena.widgets import SubWindow
import re


def write_model_by_title(
    win: SubWindow,
    dirname: str | Path,
    plugin: str | None = None,
    prefix: str = "",
) -> Path:
    """Write the widget data to a file, return the saved file."""
    model = win.to_model()
    dirname = Path(dirname)
    save_path = _get_save_path(model, dirname, prefix)

    # NOTE: default save path should not be updated, because the file is supposed to
    # be saved
    io_utils.write(model, save_path, plugin=plugin)

    i_win = int(save_path.name.split("_", 1)[0])
    _save_metadata(model.metadata, dirname, f"{i_win}_{win.title}")
    return save_path


def _get_save_path(model: WidgetDataModel, dirname: Path, prefix: str = "") -> Path:
    title = model.title or "Untitled"
    if Path(title).suffix in model.extensions:
        filename_stem = title
    else:
        if model.extension_default is None:
            if model.extensions:
                ext = model.extensions[0]
            else:
                raise ValueError(
                    f"Could not determine the file extension to be used to save {model!r}"
                )
        else:
            ext = model.extension_default
        if title.endswith(ext):
            filename_stem = title
        else:
            filename_stem = f"{title}{ext}"
    filename = f"{prefix}_{replace_invalid_characters(filename_stem)}"
    return dirname / filename


def write_metadata_by_title(
    win: SubWindow,
    dirname: str | Path,
    prefix: str = "",
) -> None:
    """Write model metadata to the default place."""
    model = win.to_model()
    dirname = Path(dirname)
    return _save_metadata(model.metadata, dirname, f"{prefix}_{win.title}")


def _save_metadata(metadata, dirname: Path, title: str):
    if isinstance(metadata, BaseMetadata):
        meta_dir = dirname / f"{title}.himena-meta"
        meta_dir.mkdir(exist_ok=True)
        write_metadata(metadata, meta_dir)
    return None


PATTERN_NOT_ALLOWED = re.compile(r"[\\/:*?\"<>|]")


def replace_invalid_characters(title: str) -> str:
    return PATTERN_NOT_ALLOWED.sub("_", title)



from typing import TYPE_CHECKING
from himena.standards.model_meta import DictMeta
from himena.plugins import register_function, configure_submenu
from himena.types import WidgetDataModel
from himena.consts import StandardType, MenuId
from himena.core import create_model

if TYPE_CHECKING:
    import numpy as np

configure_submenu(MenuId.TOOLS_EXCEL, group="20_builtins", order=20)


@register_function(
    title="Duplicate this tab",
    types=StandardType.DICT,
    menus=[MenuId.TOOLS_EXCEL],
    command_id="builtins:dict:duplicate-tab",
)
def duplicate_this_tab(
    model: WidgetDataModel[dict[str, "np.ndarray"]],
) -> WidgetDataModel["np.ndarray"]:
    """Convert the current tab into a separate window"""
    meta, tab = _meta_and_sheet(model)
    if model.type.startswith("dict."):
        type_out = model.type[5:]
    else:
        type_out = model.type
    return create_model(
        model.value[tab],
        title=f"{model.title} ({tab})",
        type=type_out,
        extension_default=".csv",
        metadata=meta.child_meta[tab],
    )


def _meta_and_sheet(model: WidgetDataModel) -> tuple[DictMeta, str]:
    if not isinstance(meta := model.metadata, DictMeta):
        raise ValueError("The input model is not a DictMeta.")
    if (tab := meta.current_tab) is None:
        raise ValueError("The current tab is not specified.")
    return meta, tab



from himena.plugins import register_dock_widget_action, add_default_status_tip
from himena.consts import MenuId

add_default_status_tip(
    short="full workflow",
    long="The full workflow widget shows the entire workflow tree of the application.",
)


@register_dock_widget_action(
    title="Full Workflow",
    menus=[MenuId.TOOLS_DOCK, MenuId.CORNER],
    area="left",
    singleton=True,
    command_id="builtins:full-workflow",
    icon="clarity:flow-chart-line",
)
def install_command_history(ui):
    """A dock widget that show the full workflow tree in the application."""
    from himena_builtins.qt.full_workflow._widget import QFullWorkflowView

    return QFullWorkflowView(ui)



from __future__ import annotations
from pathlib import Path
from qtpy import QtWidgets as QtW
from qtpy import QtGui, QtCore
from magicgui.widgets import show_file_dialog
from magicgui.widgets.bases import ValueWidget
from magicgui.types import FileDialogMode, Undefined
from magicgui.backends._qtpy.widgets import QBaseValueWidget
from magicgui.application import use_app

_MULTI_FILE_SEPARATOR = ";"


class QFileEdit(QtW.QWidget):
    valueChanged = QtCore.Signal(str)

    def __init__(self, mode=FileDialogMode.EXISTING_FILE, filter=None, parent=None):
        super().__init__(parent)
        self.setAcceptDrops(True)

        self._line_edit = QtW.QLineEdit(self)
        self._btn = QtW.QPushButton("...", self)
        self._btn.setToolTip("Browse file(s) ...")
        self._line_edit.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Preferred
        )
        self._mode = FileDialogMode(mode)
        self._filter = filter
        self._line_edit.setPlaceholderText("Enter, select or drop file(s)")

        layout = QtW.QHBoxLayout(self)
        layout.addWidget(self._line_edit)
        layout.addWidget(self._btn)
        layout.setContentsMargins(0, 0, 0, 0)

        self._btn.clicked.connect(self._open_file_dialog)

    def value(self) -> Path | list[Path] | None:
        _p = self._line_edit.text()
        if _p:
            if self._mode is FileDialogMode.EXISTING_FILES:
                return [
                    Path(p).expanduser().absolute()
                    for p in _p.split(_MULTI_FILE_SEPARATOR)
                ]
            else:
                return Path(_p.split(_MULTI_FILE_SEPARATOR)[0]).expanduser().absolute()
        return None

    def set_value(self, value: Path | list[Path] | None):
        if value is None:
            self._line_edit.clear()
        elif isinstance(value, list):
            if self._mode is not FileDialogMode.EXISTING_FILES:
                raise ValueError("Cannot set multiple files in this file mode")
            self._line_edit.setText(_MULTI_FILE_SEPARATOR.join(str(p) for p in value))
        else:
            self._line_edit.setText(str(value))

    def _open_file_dialog(self):
        _start_path = None
        if _p := self._line_edit.text():
            _fp = Path(_p.split(_MULTI_FILE_SEPARATOR)[0]).expanduser().absolute()
            if _fp.exists():
                _start_path = str(_fp)

        match self._mode:
            case FileDialogMode.EXISTING_DIRECTORY:
                _caption = "Choose directory"
            case FileDialogMode.EXISTING_FILES:
                _caption = "Select files"
            case _:
                _caption = "Select file"

        if file_path := show_file_dialog(
            self._mode,
            caption=_caption,
            start_path=_start_path,
            filter=self._filter,
            parent=self,
        ):
            if isinstance(file_path, list):
                file_path = _MULTI_FILE_SEPARATOR.join(file_path)
            self._line_edit.setText(file_path)

    def _accept_urls(self, urls: list[QtCore.QUrl]) -> bool:
        if len(urls) == 1:
            path = Path(urls[0].toLocalFile())
            if (
                self._mode is FileDialogMode.EXISTING_FILE
                and path.is_file()
                or self._mode is FileDialogMode.EXISTING_DIRECTORY
                and path.is_dir()
                or self._mode is FileDialogMode.EXISTING_FILES
                and path.is_file()
                or self._mode is FileDialogMode.OPTIONAL_FILE
                and path.is_file()
            ):
                return True
        else:
            if self._mode is FileDialogMode.EXISTING_FILES:
                return True
        return False

    def _set_urls(self, urls: list[QtCore.QUrl]):
        if urls:
            self._line_edit.setText(
                _MULTI_FILE_SEPARATOR.join(url.toLocalFile() for url in urls)
            )

    def dragEnterEvent(self, event: QtGui.QDragEnterEvent):
        if self._accept_urls(event.mimeData().urls()):
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event: QtGui.QDropEvent):
        self._set_urls(event.mimeData().urls())


class QBaseFileEdit(QBaseValueWidget):
    def __init__(self, mode: FileDialogMode, filter: str, parent=None) -> None:
        super().__init__(QFileEdit, "value", "set_value", "valueChanged", parent=parent)
        self._qwidget._mode = FileDialogMode(mode)
        self._qwidget._filter = filter


class FileEdit(ValueWidget["Path | list[Path]"]):
    def __init__(
        self,
        mode: FileDialogMode = FileDialogMode.EXISTING_FILE,
        filter: str | None = None,
        value=Undefined,
        **kwargs,
    ):
        app = use_app()
        assert app.native
        ValueWidget.__init__(
            self,
            value=value,
            widget_type=QBaseFileEdit,
            backend_kwargs={"mode": mode, "filter": filter},
            **kwargs,
        )



"""Type conversion rules."""

from typing import Literal
from io import StringIO
import csv
import numpy as np
from himena.plugins import configure_gui, register_conversion_rule, when_reader_used
from himena.types import Parametric, WidgetDataModel
from himena.consts import StandardType
from himena.core import (
    create_dataframe_model,
    create_model,
    create_text_model,
    create_array_model,
    create_table_model,
)
from himena.standards.model_meta import TableMeta
from himena.standards.roi import (
    Roi2D,
    RectangleRoi,
    PointRoi2D,
    RoiListModel,
    default_roi_label,
)
from himena.data_wrappers import wrap_dataframe, read_csv, wrap_array
from himena.utils.misc import table_to_text as _table_to_text
from himena.utils.html import html_to_plain_text


@register_conversion_rule(
    type_from=StandardType.TEXT,
    type_to=StandardType.TABLE,
    command_id="builtins:text-to-table",
)
def text_to_table(model: WidgetDataModel[str]) -> WidgetDataModel["np.ndarray"]:
    """Convert text to a table-type widget."""
    buf = StringIO(model.value)
    try:
        dialect = csv.Sniffer().sniff(buf.read(1024), delimiters=(",", "\t", ";", " "))
    except csv.Error:
        # this error is raised when there is no delimiters.
        dialect = csv.get_dialect("unix")
    sep = dialect.delimiter
    buf.seek(0)
    table = np.array(list(csv.reader(buf, delimiter=sep, quotechar=dialect.quotechar)))
    return create_table_model(
        table,
        title=model.title,
        separator=sep,
        extension_default=".csv",
    )


@register_conversion_rule(
    type_from=StandardType.TEXT,
    type_to=StandardType.ARRAY,
    command_id="builtins:text-to-array",
)
def text_to_array(model: WidgetDataModel[str]) -> WidgetDataModel["np.ndarray"]:
    """Convert text to an array-type widget using numpy."""
    text = model.value
    arr = np.loadtxt(StringIO(text), delimiter=",")
    return create_array_model(
        arr,
        title=model.title,
        extension_default=".npy",
    )


@register_conversion_rule(
    type_from=StandardType.TEXT,
    type_to=StandardType.DATAFRAME,
    command_id="builtins:text-to-dataframe",
)
def text_to_dataframe(model: WidgetDataModel[str]) -> WidgetDataModel:
    """Convert text to an dataframe-type widget."""
    from io import StringIO

    buf = StringIO(model.value)
    df = read_csv("dict", buf)
    return create_dataframe_model(
        df,
        title=model.title,
        extension_default=".csv",
    )


@register_conversion_rule(
    type_from=StandardType.HTML,
    type_to=StandardType.TEXT,
    command_id="builtins:to-plain-text",
)
def to_plain_text(model: WidgetDataModel) -> WidgetDataModel:
    """Convert HTML to plain text."""
    return create_text_model(
        html_to_plain_text(model.value),
        title=model.title,
        language="Plain Text",
        extension_default=".txt",
    )


@register_conversion_rule(
    type_from=StandardType.DATAFRAME,
    type_to=StandardType.TABLE,
    command_id="builtins:dataframe-to-table",
)
def dataframe_to_table(model: WidgetDataModel) -> WidgetDataModel["np.ndarray"]:
    """Convert a table data into a DataFrame."""
    df = wrap_dataframe(model.value)
    if isinstance(meta := model.metadata, TableMeta):
        # TODO: do we need to inherit current_position and selections?
        separator = meta.separator
    else:
        separator = ","
    return create_table_model(
        np.array([df.column_names()] + df.to_list(), dtype=np.dtypes.StringDType),
        title=model.title,
        extension_default=".csv",
        separator=separator,
    )


@register_conversion_rule(
    type_from=StandardType.DATAFRAME,
    type_to=StandardType.TEXT,
    command_id="builtins:dataframe-to-text",
)
def dataframe_to_text(model: WidgetDataModel) -> Parametric:
    """Convert a table data into a DataFrame."""

    def convert_dataframe_to_text(
        format: Literal["CSV", "TSV", "Markdown", "Latex", "rST", "HTML"] = "CSV",
        end_of_text: Literal["", "\\n"] = "\\n",
    ) -> WidgetDataModel[str]:
        df = wrap_dataframe(model.value)
        table_input = df.to_list()
        table_input.insert(0, df.column_names())
        end_of_text = "\n" if end_of_text == "\\n" else ""
        value, ext_default, language = _table_to_text(table_input, format, end_of_text)
        return create_text_model(
            value,
            title=model.title,
            extension_default=ext_default,
            language=language,
        )

    return convert_dataframe_to_text


@register_conversion_rule(
    type_from=StandardType.TABLE,
    type_to=StandardType.TEXT,
    command_id="builtins:table-to-text",
)
def table_to_text(model: WidgetDataModel) -> Parametric:
    """Convert a table data into a text data."""

    @configure_gui(
        preview=True, end_of_text={"choices": [("none", ""), ("newline", "\n")]}
    )
    def convert_table_to_text(
        format: Literal["CSV", "TSV", "Markdown", "Latex", "rST", "HTML"] = "CSV",
        end_of_text: Literal["", "\n"] = "\n",
    ) -> WidgetDataModel[str]:
        value, ext_default, language = _table_to_text(model.value, format, end_of_text)
        return create_text_model(
            value,
            title=model.title,
            extension_default=ext_default,
            language=language,
        )

    return convert_table_to_text


@register_conversion_rule(
    type_from=StandardType.TABLE,
    type_to=StandardType.DATAFRAME,
    command_id="builtins:table-to-dataframe",
)
def table_to_dataframe(model: WidgetDataModel["np.ndarray"]) -> WidgetDataModel:
    """Convert a table data into a DataFrame."""
    # TODO: Choose pandas, dict etc.
    if isinstance(meta := model.metadata, TableMeta):
        separator = meta.separator
    else:
        separator = ","
    buf = StringIO()
    np.savetxt(buf, model.value, fmt="%s", delimiter=",")
    buf.seek(0)
    df = read_csv("dict", buf)
    return create_dataframe_model(
        df,
        title=model.title,
        extension_default=".csv",
        separator=separator,
    )


@register_conversion_rule(
    type_from=StandardType.TABLE,
    type_to=StandardType.ARRAY,
    command_id="builtins:table-to-array",
)
def table_to_array(model: WidgetDataModel) -> WidgetDataModel:
    """Convert a table data into an array."""
    arr_str = model.value

    def _try_astype(arr_str, dtype) -> tuple["np.ndarray", bool]:
        if not isinstance(arr_str, np.ndarray):
            arr_str = np.array(arr_str, dtype=np.dtypes.StringDType())
        try:
            arr = arr_str.astype(dtype)
            ok = True
        except ValueError:
            arr = arr_str
            ok = False
        return arr, ok

    arr, ok = _try_astype(arr_str, int)
    if not ok:
        arr, ok = _try_astype(arr_str, float)
    if not ok:
        arr, ok = _try_astype(arr_str, complex)
    if not ok:
        pass

    return create_array_model(
        arr,
        title=model.title,
        extension_default=".npy",
    )


@register_conversion_rule(
    type_from=StandardType.DATAFRAME,
    type_to=StandardType.ROIS,
    command_id="builtins:dataframe-to-image-rois",
)
def dataframe_to_image_rois(model: WidgetDataModel) -> Parametric:
    """Convert a data frame data into image ROIs."""

    @configure_gui(roi_type={"choices": ["rectangle", "point"]})
    def convert_dataframe_to_image_rois(
        roi_type: str,
        indices: list[str] = (),
    ) -> WidgetDataModel:
        df = wrap_dataframe(model.value)
        if indices:
            arr_indice = np.stack(
                [df.column_to_array(indice_column) for indice_column in indices],
                axis=1,
                dtype=np.int32,
            )
        else:
            arr_indice = np.empty((len(df), 0), dtype=np.int32)
        rois: list[Roi2D] = []
        if roi_type == "rectangle":
            arr_xywh = np.stack(
                [
                    df.column_to_array("x"),
                    df.column_to_array("y"),
                    df.column_to_array("width"),
                    df.column_to_array("height"),
                ],
                axis=1,
            )
            for idx, xywh in enumerate(arr_xywh):
                x, y, w, h = xywh
                rois.append(
                    RectangleRoi(
                        name=default_roi_label(idx),
                        x=x,
                        y=y,
                        width=w,
                        height=h,
                    )
                )
        elif roi_type == "point":
            arr_xywh = np.stack(
                [
                    df.column_to_array("x"),
                    df.column_to_array("y"),
                ],
                axis=1,
            )
            for idx, xy in enumerate(arr_xywh):
                x, y = xy
                rois.append(PointRoi2D(name=default_roi_label(idx), x=x, y=y))
        else:
            raise ValueError("Only 'rectangle' and 'point' are supported.")
        return create_model(
            RoiListModel(items=rois, indices=arr_indice, axis_names=indices),
            title=model.title,
            type=StandardType.ROIS,
        )

    return convert_dataframe_to_image_rois


@register_conversion_rule(
    type_from=StandardType.DATAFRAME,
    type_to=StandardType.DATAFRAME_PLOT,
    command_id="builtins:dataframe-to-dataframe-plot",
)
def dataframe_to_dataframe_plot(model: WidgetDataModel) -> WidgetDataModel:
    df = wrap_dataframe(model.value)
    col_non_numerical = [
        col
        for col, dtype in zip(df.column_names(), df.dtypes)
        if dtype.kind not in "uifb"
    ]
    if col_non_numerical:
        raise ValueError(
            f"DataFrame contains non-numerical value in columns {col_non_numerical!r}"
        )
    return model.astype(StandardType.DATAFRAME_PLOT)


@register_conversion_rule(
    type_from=StandardType.ARRAY,
    type_to=StandardType.TABLE,
    command_id="builtins:array-to-table",
)
def array_to_table(model: WidgetDataModel) -> WidgetDataModel:
    """Convert an array data into a table."""
    arr = wrap_array(model.value)
    if arr.ndim > 2:
        raise ValueError("Cannot convert >2D array to a table.")
    value = arr.get_slice(()).astype(np.dtypes.StringDType())
    if value.ndim < 2:
        value = np.atleast_2d(value)
    return create_table_model(
        value,
        title=model.title,
        extension_default=".csv",
    )


when_reader_used(StandardType.TABLE).add_command_suggestion(
    "builtins:table-to-dataframe"
)



from __future__ import annotations
from typing import Iterable
import weakref
from qtpy.QtCore import QTimer
from himena.plugins import validate_protocol
from himena.widgets import MainWindow, SubWindow
import himena.workflow as _wf
from himena_builtins.qt.widgets.workflow import QWorkflowViewBase, WorkflowNodeItem


class QFullWorkflowView(QWorkflowViewBase):
    def __init__(self, ui: MainWindow):
        super().__init__()
        self._ui_ref = weakref.ref(ui)
        self.view.item_left_clicked.connect(self._find_window)
        self._single_shot = QTimer.singleShot  # just for testing

    @validate_protocol
    def widget_added_callback(self):
        if ui := self._ui_ref():
            for win in ui.iter_windows():
                self.add_workflow(win._widget_workflow)
            ui.events.window_added.connect(self._window_added)
            ui.events.window_closed.connect(self._window_closed)
            ui.events.window_activated.connect(self._window_activated)

    def _window_added(self, win: SubWindow):
        self._single_shot(0, lambda win=win: self.add_window(win))

    def _window_closed(self, win_closed: SubWindow):
        if ui := self._ui_ref():
            workflows: list[_wf.Workflow] = []
            for win in ui.iter_windows():
                if win is win_closed:
                    continue
                workflows.append(win._widget_workflow)
            self.reset_workflows(workflows)

    def _window_activated(self, win: SubWindow):
        if node := self.view._node_map.get(win._identifier):
            node.setSelected(True)

    def _find_window(self, item: WorkflowNodeItem) -> None:
        if ui := self._ui_ref():
            step = item._step
            for i_tab, tab in ui.tabs.enumerate():
                for i_win, win in tab.enumerate():
                    if win._identifier == step.id:
                        ui.tabs.current_index = i_tab
                        ui.tabs[i_tab].current_index = i_win
                        return None

    def add_window(self, window: SubWindow) -> None:
        """Add a window's workflow to the view."""
        if not window._widget_workflow:
            return
        self.add_workflow(window._widget_workflow)

    def reset_workflows(self, workflows: Iterable[_wf.Workflow]) -> None:
        """Remove the workflow from the view."""
        not_visited = self.view._node_map.copy()
        for workflow in workflows:
            for step in workflow:
                not_visited.pop(step.id, None)
        self.view.remove_nodes(not_visited.values())



from __future__ import annotations

from qtpy import QtWidgets as QtW
from himena.plugins import validate_protocol
from himena.qt._utils import drag_command
from himena.standards import roi
from himena.standards.model_meta import ImageRoisMeta, DimAxis
from himena.consts import StandardType
from himena.types import WidgetDataModel
from ._image_components import QSimpleRoiCollection


class QImageRoiView(QtW.QWidget):
    """The ROI list widget"""

    __himena_widget_id__ = "builtins:QImageRoiView"
    __himena_display_name__ = "Built-in Image ROI Viewer"

    def __init__(self):
        super().__init__()

        layout = QtW.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self._roi_collection = QSimpleRoiCollection()
        layout.addWidget(self._roi_collection)
        self._is_modified = False
        self._model_type = StandardType.ROIS
        self._axes: list[DimAxis] | None = None
        self._roi_collection.drag_requested.connect(self._on_drag_requested)

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        value = model.value
        if callable(value):
            value = value()
        if not isinstance(value, roi.RoiListModel):
            raise ValueError(f"Expected a RoiListModel, got {type(value)}")
        self._roi_collection.clear()
        self._roi_collection.extend_from_standard_roi_list(value)
        self._model_type = model.type
        if isinstance(meta := model.metadata, ImageRoisMeta):
            if meta.axes:
                self._axes = meta.axes
            if meta.selections:
                self._roi_collection.set_selections(meta.selections)

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        return WidgetDataModel(
            value=self._roi_collection.to_standard_roi_list(),
            type=self.model_type(),
            extension_default=".roi.json",
            metadata=ImageRoisMeta(
                axes=self._axes,
                selections=self._roi_collection.selections(),
            ),
        )

    @validate_protocol
    def model_type(self) -> str:
        return self._model_type

    @validate_protocol
    def is_modified(self) -> bool:
        return self._is_modified

    @validate_protocol
    def set_modified(self, modified: bool):
        self._is_modified = modified

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return 180, 300

    def _on_drag_requested(self, indices: list[int]):
        nrois = len(indices)
        _s = "" if nrois == 1 else "s"
        return drag_command(
            self,
            "builtins:rois:select",
            StandardType.ROIS,
            with_params={"selections": indices},
            desc=f"{nrois} ROI{_s}",
        )



from __future__ import annotations
from contextlib import contextmanager
import math
from enum import Enum, auto
from qtpy import QtWidgets as QtW, QtCore, QtGui
from psygnal import Signal
import numpy as np
from typing import Iterable, Iterator, TYPE_CHECKING

from himena.standards import roi

if TYPE_CHECKING:
    from typing import Self


class QRoi(QtW.QGraphicsItem):
    """The base class for all ROI items."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.label()!r} at {hex(id(self))}>"

    def label(self) -> str:
        return getattr(self, "_roi_label", "")

    def set_label(self, label: str):
        self._roi_label = label

    def toRoi(self) -> roi.RoiModel:
        raise NotImplementedError

    def translate(self, dx: float, dy: float):
        raise NotImplementedError

    def makeThumbnail(self, pixmap: QtGui.QPixmap) -> QtGui.QPixmap:
        raise NotImplementedError

    def withPen(self, pen: QtGui.QPen):
        self.setPen(pen)
        return self

    def withLabel(self, label: str | None):
        self.set_label(label or "")
        return self

    def copy(self) -> Self:
        raise NotImplementedError("Subclasses must implement this method.")

    def _thumbnail_transform(self, width: int, height: int) -> QtGui.QTransform:
        rect = self.boundingRect()
        transform = QtGui.QTransform()
        rect_size = max(rect.width(), rect.height())
        if rect_size == 0:
            rect_size = 1
        transform.translate(width / 2, height / 2)
        transform.scale((width - 2) / rect_size, (height - 2) / rect_size)
        transform.translate(-rect.center().x(), -rect.center().y())
        return transform

    def _roi_type(self) -> str:
        return self.__class__.__name__

    def short_description(self, xscale: float, yscale: float, unit: str) -> str:
        """Description that will be shown in the tooltip when the ROI is drawn."""
        return ""


class QLineRoi(QtW.QGraphicsLineItem, QRoi):
    changed = Signal(QtCore.QLineF)

    def toRoi(self) -> roi.LineRoi:
        x1, y1, x2, y2 = self._coordinates()
        return roi.LineRoi(
            start=(x1 - 0.5, y1 - 0.5),
            end=(x2 - 0.5, y2 - 0.5),
            name=self.label(),
        )

    def _coordinates(self) -> tuple[float, float, float, float]:
        line = self.line()
        return line.x1(), line.y1(), line.x2(), line.y2()

    def translate(self, dx: float, dy: float):
        new_line = self.line()
        new_line.translate(dx, dy)
        self.setLine(new_line)

    def setLine(self, *args):
        super().setLine(*args)
        self.changed.emit(self.line())

    def makeThumbnail(self, pixmap: QtGui.QPixmap) -> QtGui.QPixmap:
        painter = QtGui.QPainter(pixmap)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        painter.setPen(self.pen())
        painter.setTransform(self._thumbnail_transform(pixmap.width(), pixmap.height()))
        painter.drawLine(self.line())
        painter.end()
        return pixmap

    def copy(self) -> QLineRoi:
        return QLineRoi(self.line()).withPen(self.pen())

    def _roi_type(self) -> str:
        return "line"

    def short_description(self, xscale: float, yscale: float, unit: str) -> str:
        x1, y1, x2, y2 = self._coordinates()
        start = f"{x1:.1f}, {y1:.1f}"
        end = f"{x2:.1f}, {y2:.1f}"
        dx, dy = x2 - x1, y2 - y1
        length_px = math.sqrt(dx**2 + dy**2)
        ang_px = self.toRoi().angle()
        if not unit:
            return f"start=({start}), end=({end}), length={length_px:.1f}, angle={ang_px:.1f}°"
        length = math.sqrt((dx * xscale) ** 2 + (dy * yscale) ** 2)
        ang = -math.degrees(math.atan2(dy * yscale, dx * xscale))
        return f"start=({start})<br>end=({end})<br>length={length_px:.1f} px ({length:.1f} {unit})<br>angle={ang_px:.1f}° (scaled: {ang:.1f}°)"


class QRectRoiBase(QRoi):
    changed = Signal(QtCore.QRectF)


class QRectangleRoi(QtW.QGraphicsRectItem, QRectRoiBase):
    def toRoi(self) -> roi.RectangleRoi:
        rect = self.rect()
        x, y, width, height = _may_be_ints(
            rect.x(), rect.y(), rect.width(), rect.height()
        )
        return roi.RectangleRoi(
            x=x,
            y=y,
            width=width,
            height=height,
            name=self.label(),
        )

    def setRect(self, *args):
        super().setRect(*args)
        self.changed.emit(self.rect())

    def translate(self, dx: float, dy: float):
        new_rect = self.rect()
        new_rect.translate(dx, dy)
        self.setRect(new_rect)

    def makeThumbnail(self, pixmap: QtGui.QPixmap) -> QtGui.QPixmap:
        painter = QtGui.QPainter(pixmap)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        painter.setPen(self.pen())
        painter.setTransform(self._thumbnail_transform(pixmap.width(), pixmap.height()))
        painter.drawRect(self.rect())
        painter.end()
        return pixmap

    def copy(self) -> QRectangleRoi:
        return QRectangleRoi(self.rect()).withPen(self.pen())

    def _roi_type(self) -> str:
        return "rectangle"

    def short_description(self, xscale: float, yscale: float, unit: str) -> str:
        rect = self.rect()
        x, y, w, h = rect.x(), rect.y(), rect.width(), rect.height()
        width_px = w
        height_px = h
        if not unit:
            return f"left={x:.1f}<br>top={y:.1f}<br>width={width_px:.1f}<br>height={height_px:.1f}"
        width = w * xscale
        height = h * yscale
        return f"left={x:.1f}<br>top={y:.1f}<br>width={width_px:.1f} ({width:.1f} {unit})<br>height={height_px:.1f} ({height:.1f} {unit})"


class QEllipseRoi(QtW.QGraphicsEllipseItem, QRectRoiBase):
    changed = Signal(QtCore.QRectF)

    def toRoi(self) -> roi.EllipseRoi:
        rect = self.rect()
        x, y, width, height = _may_be_ints(
            rect.x(), rect.y(), rect.width(), rect.height()
        )
        return roi.EllipseRoi(
            x=x,
            y=y,
            width=width,
            height=height,
            name=self.label(),
        )

    def setRect(self, *args):
        super().setRect(*args)
        self.changed.emit(self.rect())

    def translate(self, dx: float, dy: float):
        new_rect = self.rect()
        new_rect.translate(dx, dy)
        self.setRect(new_rect)

    def makeThumbnail(self, pixmap: QtGui.QPixmap) -> QtGui.QPixmap:
        painter = QtGui.QPainter(pixmap)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        painter.setPen(self.pen())
        painter.setTransform(self._thumbnail_transform(pixmap.width(), pixmap.height()))
        painter.drawEllipse(self.rect())
        painter.end()
        return pixmap

    def copy(self) -> QEllipseRoi:
        return QEllipseRoi(self.rect()).withPen(self.pen())

    def _roi_type(self) -> str:
        return "ellipse"

    def short_description(self, xscale: float, yscale: float, unit: str) -> str:
        rect = self.rect()
        x, y, w, h = rect.x(), rect.y(), rect.width(), rect.height()
        cx = x + w / 2
        cy = y + h / 2
        width = w * xscale
        height = h * yscale
        if w == 0 or h == 0:
            eccentricity = "n.a."
        elif w < h:
            eccentricity = format(math.sqrt(1 - (width / height) ** 2), ".2f")
        else:
            eccentricity = format(math.sqrt(1 - (height / width) ** 2), ".2f")
        if not unit:
            return f"center=[{cx:.1f}, {cy:.1f}]<br>width={w:.1f}<br>height={h:.1f}<br>eccentricity={eccentricity}"
        return f"center=[{cx:.1f}, {cy:.1f}]<br>width={w:.1f} ({width:.1f} {unit})<br>height={h:.1f} ({height:.1f} {unit})<br>eccentricity={eccentricity}"


class QRotatedRoiBase(QRoi):
    """Rotated ROI base class.

    A rotated ROI looks like this:

    ```
           2             1: start
      +----o----+        2: top
    1 o         o 4      3: bottom
      +----o----+        4: end
           3             Drawn in the 1 -> 4 direction.
    ```

    """

    changed = Signal(object)

    def __init__(self, start: QtCore.QPointF, end: QtCore.QPointF, width: float = 50):
        super().__init__()
        dr = end - start
        length = math.sqrt(dr.x() ** 2 + dr.y() ** 2)
        center = (start + end) / 2
        rad = math.atan2(dr.y(), dr.x())
        self._center = center
        self._angle = math.degrees(rad)
        self._length = length
        self._width = width
        self._pen = QtGui.QPen()
        self.set_start(start)
        self.set_end(end)

    def vector_x(self) -> QtCore.QPointF:
        rad = math.radians(self.angle())
        return QtCore.QPointF(math.cos(rad), math.sin(rad)) * self._length

    def vector_y(self) -> QtCore.QPointF:
        rad = math.radians(self.angle())
        return QtCore.QPointF(-math.sin(rad), math.cos(rad)) * self._width

    def angle(self) -> float:
        return self._angle

    def start(self) -> QtCore.QPointF:
        """Return the left anchor point."""
        return self._center - self.vector_x() / 2

    def end(self) -> QtCore.QPointF:
        """Return the right anchor point."""
        return self._center + self.vector_x() / 2

    def top(self) -> QtCore.QPointF:
        """Return the top anchor point."""
        return self._center - self.vector_y() / 2

    def bottom(self) -> QtCore.QPointF:
        """Return the bottom anchor point."""
        return self._center + self.vector_y() / 2

    def center(self) -> QtCore.QPointF:
        return self._center

    def set_start(self, left: QtCore.QPointF):
        right = self.end()
        vecx = right - left
        with self._update_and_emit():
            self._angle = math.degrees(math.atan2(vecx.y(), vecx.x()))
            self._length = math.sqrt(vecx.x() ** 2 + vecx.y() ** 2)
            self._center = (left + right) / 2

    def set_end(self, right: QtCore.QPointF):
        left = self.start()
        vecx = right - left
        with self._update_and_emit():
            self._angle = math.degrees(math.atan2(vecx.y(), vecx.x()))
            self._length = math.sqrt(vecx.x() ** 2 + vecx.y() ** 2)
            self._center = (left + right) / 2

    def set_width(self, width: float):
        with self._update_and_emit():
            self._width = width

    def pen(self) -> QtGui.QPen:
        return self._pen

    def setPen(self, pen: QtGui.QPen):
        self._pen = QtGui.QPen(pen)
        self.update()

    def translate(self, dx: float, dy: float):
        with self._update_and_emit():
            self._center += QtCore.QPointF(dx, dy)

    def copy(self) -> Self:
        cls = type(self)
        return cls(self.start(), self.end(), self._width).withPen(self.pen())

    @contextmanager
    def _update_and_emit(self):
        old_bbox = self.boundingRect()
        yield
        new_bbox = self.boundingRect()
        self.changed.emit(self)
        self.update()
        if scene := self.scene():
            scene.update(self.mapRectToScene(old_bbox.united(new_bbox)))

    def short_description(self, xscale: float, yscale: float, unit: str) -> str:
        start = self.start()
        end = self.end()
        length_px = self._length
        if not unit:
            return f"start=[{start.x():.1f}, {start.y():.1f}]<br>end=[{end.x():.1f}, {end.y():.1f}]<br>length={length_px:.1f}<br>angle={self.angle():.1f}°"
        length = length_px * xscale
        return f"start=[{start.x():.1f}, {start.y():.1f}]<br>end=[{end.x():.1f}, {end.y():.1f}]<br>length={length_px:.1f} ({length:.1f} {unit})<br>angle={self.angle():.1f}°"


class QRotatedRectangleRoi(QRotatedRoiBase):
    def toRoi(self) -> roi.RotatedRectangleRoi:
        return roi.RotatedRectangleRoi(
            start=(self.start().x() - 0.5, self.start().y() - 0.5),
            end=(self.end().x() - 0.5, self.end().y() - 0.5),
            width=self._width,
            name=self.label(),
        )

    def _corner_points(self) -> list[QtCore.QPointF]:
        center = self.center()
        vx = self.vector_x()
        vy = self.vector_y()
        p00 = center - vx / 2 - vy / 2
        p01 = center - vx / 2 + vy / 2
        p10 = center + vx / 2 - vy / 2
        p11 = center + vx / 2 + vy / 2
        return p00, p01, p11, p10

    def paint(self, painter, option, widget):
        painter.setPen(self.pen())
        painter.drawPolygon(*self._corner_points())

    def boundingRect(self):
        points = self._corner_points()
        xmin = min(p.x() for p in points)
        xmax = max(p.x() for p in points)
        ymin = min(p.y() for p in points)
        ymax = max(p.y() for p in points)
        return QtCore.QRectF(xmin, ymin, xmax - xmin, ymax - ymin)

    def contains(self, point: QtCore.QPointF) -> bool:
        polygon = QtGui.QPolygonF(self._corner_points())
        return polygon.containsPoint(point, QtCore.Qt.FillRule.WindingFill)

    def makeThumbnail(self, pixmap: QtGui.QPixmap) -> QtGui.QPixmap:
        painter = QtGui.QPainter(pixmap)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        painter.setPen(self.pen())
        painter.setTransform(self._thumbnail_transform(pixmap.width(), pixmap.height()))
        painter.drawPolygon(*self._corner_points())
        painter.end()
        return pixmap

    def _roi_type(self) -> str:
        return "rotated rectangle"


class QRotatedEllipseRoi(QRotatedRoiBase):
    def toRoi(self) -> roi.RotatedEllipseRoi:
        return roi.RotatedEllipseRoi(
            start=(self.start().x() - 0.5, self.start().y() - 0.5),
            end=(self.end().x() - 0.5, self.end().y() - 0.5),
            width=self._width,
            name=self.label(),
        )

    def _point_array(self) -> np.ndarray:
        """(N, 2) array of points that represent the ellipse."""
        a = self._length / 2
        b = self._width / 2
        rad_ellipse = math.radians(self.angle())
        cos = math.cos(rad_ellipse)
        sin = math.sin(rad_ellipse)
        rads = np.linspace(0, 2 * np.pi, 60, endpoint=False)
        x = self._center.x() + a * np.cos(rads) * cos - b * np.sin(rads) * sin
        y = self._center.y() + a * np.cos(rads) * sin + b * np.sin(rads) * cos
        return np.stack([x, y], axis=1)

    def _dense_points(self) -> list[QtCore.QPointF]:
        """Return a list of points that represent the ellipse."""
        points = self._point_array()
        return [QtCore.QPointF(x, y) for x, y in points]

    def paint(self, painter, option, widget):
        painter.setPen(self.pen())
        painter.drawPolygon(*self._dense_points())

    def boundingRect(self):
        points = self._point_array()
        xmin, ymin = points.min(axis=0)
        xmax, ymax = points.max(axis=0)
        return QtCore.QRectF(xmin, ymin, xmax - xmin, ymax - ymin)

    def contains(self, point: QtCore.QPointF) -> bool:
        polygon = QtGui.QPolygonF(self._dense_points())
        return polygon.containsPoint(point, QtCore.Qt.FillRule.WindingFill)

    def makeThumbnail(self, pixmap: QtGui.QPixmap) -> QtGui.QPixmap:
        painter = QtGui.QPainter(pixmap)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        painter.setPen(self.pen())
        painter.setTransform(self._thumbnail_transform(pixmap.width(), pixmap.height()))
        painter.drawPolygon(*self._dense_points())
        painter.end()
        return pixmap

    def _roi_type(self) -> str:
        return "rotated ellipse"


class QSegmentedLineRoi(QtW.QGraphicsPathItem, QRoi):
    changed = Signal(QtGui.QPainterPath)

    def __init__(self, xs: Iterable[float], ys: Iterable[float], parent=None):
        super().__init__(parent)
        path = QtGui.QPainterPath()
        path.moveTo(xs[0], ys[0])
        for x, y in zip(xs[1:], ys[1:]):
            path.lineTo(x, y)
        self.setPath(path)

    def toRoi(self) -> roi.SegmentedLineRoi:
        path = self.path()
        xs, ys = [], []
        for i in range(path.elementCount()):
            element = path.elementAt(i)
            xs.append(element.x - 0.5)
            ys.append(element.y - 0.5)
        return roi.SegmentedLineRoi(xs=xs, ys=ys, name=self.label())

    def setPath(self, *args):
        super().setPath(*args)
        self.changed.emit(self.path())

    def translate(self, dx: float, dy: float):
        new_path = self.path()
        new_path.translate(dx, dy)
        self.setPath(new_path)

    def add_point(self, pos: QtCore.QPointF):
        path = self.path()
        if path.elementCount() == 0:
            path.moveTo(pos)
        else:
            path.lineTo(pos)
        self.setPath(path)

    def count(self) -> int:
        """Number of points in the line."""
        return self.path().elementCount()

    def update_point(self, ith: int, pos: QtCore.QPointF):
        path = self.path()
        path.setElementPositionAt(ith, pos.x(), pos.y())
        self.setPath(path)

    def makeThumbnail(self, pixmap: QtGui.QPixmap) -> QtGui.QPixmap:
        painter = QtGui.QPainter(pixmap)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        painter.setPen(self.pen())
        painter.setTransform(self._thumbnail_transform(pixmap.width(), pixmap.height()))
        painter.drawPath(self.path())
        painter.end()
        return pixmap

    def copy(self) -> QSegmentedLineRoi:
        path = self.path()
        xs, ys = [], []
        for i in range(path.elementCount()):
            element = path.elementAt(i)
            xs.append(element.x)
            ys.append(element.y)
        return QSegmentedLineRoi(xs, ys).withPen(self.pen())

    def _roi_type(self) -> str:
        return "segmented line"


class QPolygonRoi(QSegmentedLineRoi):
    def toRoi(self) -> roi.PolygonRoi:
        path = self.path()
        xs, ys = [], []
        closed = path.elementAt(0) == path.elementAt(path.elementCount() - 1)
        for i in range(path.elementCount() - int(closed)):
            element = path.elementAt(i)
            xs.append(element.x - 0.5)
            ys.append(element.y - 0.5)
        return roi.PolygonRoi(xs=xs, ys=ys, name=self.label())

    def _roi_type(self) -> str:
        return "polygon"


class QPointRoiBase(QRoi):
    def __init__(self, parent):
        super().__init__(parent)
        self._pen = QtGui.QPen(QtGui.QColor(0, 0, 0), 2)
        self._pen.setCosmetic(True)
        self._pen.setJoinStyle(QtCore.Qt.PenJoinStyle.MiterJoin)
        self._brush = QtGui.QBrush(QtGui.QColor(225, 225, 0))
        self._size = 4.5
        symbol = QtGui.QPainterPath()
        symbol.moveTo(-self._size, 0)
        symbol.lineTo(self._size, 0)
        symbol.moveTo(0, -self._size)
        symbol.lineTo(0, self._size)
        self._symbol = symbol
        self._bounding_rect_cache: QtCore.QRectF | None = None

    def pen(self) -> QtGui.QPen:
        return self._pen

    def setPen(self, pen: QtGui.QPen):
        self._pen = pen

    def brush(self) -> QtGui.QBrush:
        return self._brush

    def setBrush(self, brush: QtGui.QBrush):
        self._brush = brush

    def _iter_points(self) -> Iterator[QtCore.QPointF]:
        raise NotImplementedError

    def _repr_points(self) -> Iterable[tuple[float, float]]:
        """Return a list of (x, y) coordinates for drawing thumbnails."""
        raise NotImplementedError

    def paint(
        self,
        painter: QtGui.QPainter,
        option: QtW.QStyleOptionGraphicsItem,
        widget: QtW.QWidget,
    ):
        painter.setPen(self.pen())
        painter.setBrush(self.brush())
        tr = painter.transform()

        for pt in self._iter_points():
            painter.resetTransform()
            xy_transformed = tr.map(pt)
            painter.translate(xy_transformed)
            painter.drawPath(self._symbol)
        self.scene().update()

    def makeThumbnail(self, pixmap: QtGui.QPixmap) -> QtGui.QPixmap:
        painter = QtGui.QPainter(pixmap)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        painter.setPen(self.pen())
        for rx, ry in self._repr_points():
            painter.resetTransform()
            painter.translate(QtCore.QPointF(pixmap.width() * rx, pixmap.height() * ry))
            painter.scale(0.3, 0.3)
            painter.drawPath(self._symbol)
        return pixmap


class QPointRoi(QPointRoiBase):
    changed = Signal(QtCore.QPointF)

    def __init__(self, x: float, y: float, parent=None):
        super().__init__(parent)
        self._point = QtCore.QPointF(x, y)

    def point(self) -> QtCore.QPointF:
        return self._point

    def _iter_points(self) -> Iterator[QtCore.QPointF]:
        yield self._point

    def setPoint(self, point: QtCore.QPointF):
        self._point = point
        self.changed.emit(self._point)

    def toRoi(self) -> roi.PointRoi2D:
        return roi.PointRoi2D(
            x=self._point.x() - 0.5,
            y=self._point.y() - 0.5,
            name=self.label(),
        )

    def translate(self, dx: float, dy: float):
        self.setPoint(QtCore.QPointF(self._point.x() + dx, self._point.y() + dy))

    def boundingRect(self) -> QtCore.QRectF:
        return QtCore.QRectF(
            self._point.x() - self._size,
            self._point.y() - self._size,
            self._size * 2,
            self._size * 2,
        )

    def copy(self) -> QPointRoi:
        return QPointRoi(self._point.x(), self._point.y()).withPen(self.pen())

    def _roi_type(self) -> str:
        return "point"

    def _repr_points(self):
        return [(0.5, 0.5)]


class QPointsRoi(QPointRoiBase):
    changed = Signal(list)

    def __init__(self, xs: Iterable[float], ys: Iterable[float], parent=None):
        super().__init__(parent)
        self._points = [QtCore.QPointF(x, y) for x, y in zip(xs, ys)]

    def count(self) -> int:
        return len(self._points)

    def pointAt(self, idx: int) -> QtCore.QPointF:
        return self._points[idx]

    def update_point(self, idx: int, pos: QtCore.QPointF):
        self._points[idx] = pos
        self.changed.emit(self._points)
        self._bounding_rect_cache = None

    def toRoi(self) -> roi.PointsRoi2D:
        xs: list[float] = []
        ys: list[float] = []
        for point in self._points:
            xs.append(point.x() - 0.5)
            ys.append(point.y() - 0.5)
        return roi.PointsRoi2D(xs=np.array(xs), ys=np.array(ys), name=self.label())

    def translate(self, dx: float, dy: float):
        self._points = [
            QtCore.QPointF(point.x() + dx, point.y() + dy) for point in self._points
        ]
        self.changed.emit(self._points)
        self._bounding_rect_cache = self._bounding_rect_cache.translated(dx, dy)

    def add_point(self, pos: QtCore.QPointF):
        self._points.append(pos)
        self.changed.emit(self._points)
        self._bounding_rect_cache = None

    def boundingRect(self) -> QtCore.QRectF:
        if self._bounding_rect_cache is None:
            xmin = np.min([pt.x() for pt in self._points])
            xmax = np.max([pt.x() for pt in self._points])
            ymin = np.min([pt.y() for pt in self._points])
            ymax = np.max([pt.y() for pt in self._points])
            self._bounding_rect_cache = QtCore.QRectF(
                xmin - self._size / 2,
                ymin - self._size / 2,
                xmax - xmin + self._size,
                ymax - ymin + self._size,
            )
        return self._bounding_rect_cache

    def _iter_points(self) -> Iterator[QtCore.QPointF]:
        yield from self._points

    def copy(self) -> QPointsRoi:
        return QPointsRoi(
            [pt.x() for pt in self._points], [pt.y() for pt in self._points]
        ).withPen(self.pen())

    def _roi_type(self) -> str:
        return "points"

    def _repr_points(self):
        return [(0.2, 0.2), (0.5, 0.8), (0.8, 0.6)]


class QCircleRoi(QtW.QGraphicsEllipseItem, QRoi):
    changed = Signal(QtCore.QRectF)

    def __init__(self, x: float, y: float, radius: float, parent=None):
        super().__init__(parent)
        self._radius = radius
        self.setRect(x - radius, y - radius, radius * 2, radius * 2)

    def toRoi(self) -> roi.CircleRoi:
        rect = self.rect()
        return roi.CircleRoi(
            x=rect.x() + rect.width() / 2 - 0.5,
            y=rect.y() + rect.height() / 2 - 0.5,
            radius=self._radius,
            name=self.label(),
        )

    def setRect(self, *args):
        super().setRect(*args)
        self.changed.emit(self.rect())

    def radius(self) -> float:
        """Return the radius of the circle."""
        return self._radius

    def center(self) -> QtCore.QPointF:
        """Return the center of the circle."""
        rect = self.rect()
        return QtCore.QPointF(rect.x() + rect.width() / 2, rect.y() + rect.height() / 2)

    def setCenterAndRadius(self, center: tuple[float, float], radius: float):
        """Set the radius of the circle."""
        self._radius = radius
        self.setRect(center[0] - radius, center[1] - radius, radius * 2, radius * 2)

    def translate(self, dx: float, dy: float):
        new_rect = self.rect()
        new_rect.translate(dx, dy)
        self.setRect(new_rect)

    def makeThumbnail(self, pixmap: QtGui.QPixmap) -> QtGui.QPixmap:
        painter = QtGui.QPainter(pixmap)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        painter.setPen(self.pen())
        painter.setTransform(self._thumbnail_transform(pixmap.width(), pixmap.height()))
        painter.drawEllipse(self.rect())
        painter.end()
        return pixmap

    def copy(self) -> QCircleRoi:
        return QCircleRoi(
            self.rect().x() + self._radius,
            self.rect().y() + self._radius,
            self._radius,
        ).withPen(self.pen())

    def _roi_type(self) -> str:
        return "circle"

    def contains(self, point: QtCore.QPointF) -> bool:
        """Check if the point is inside the circle."""
        center = self.center()
        dx = point.x() - center.x()
        dy = point.y() - center.y()
        return dx**2 + dy**2 <= self._radius**2

    def short_description(self, xscale: float, yscale: float, unit: str) -> str:
        rect = self.rect()
        cx = rect.x() + rect.width() / 2
        cy = rect.y() + rect.height() / 2
        radius_px = self._radius
        radius = radius_px * xscale
        if not unit:
            return f"center=[{cx:.1f}, {cy:.1f}]<br>radius={radius_px:.1f}"
        return f"center=[{cx:.1f}, {cy:.1f}]<br>radius={radius_px:.1f} ({radius:.1f} {unit})"


def _may_be_ints(*values: float) -> list[int | float]:
    """Convert values to int if they are close to an integer."""
    return [int(v) if abs(v - round(v)) < 1e-6 else v for v in values]


class MouseMode(Enum):
    """Mouse interaction modes for the image graphics view."""

    SELECT = auto()
    PAN_ZOOM = auto()
    ROI_RECTANGLE = auto()
    ROI_ROTATED_RECTANGLE = auto()
    ROI_ELLIPSE = auto()
    ROI_ROTATED_ELLIPSE = auto()
    ROI_POINT = auto()
    ROI_POINTS = auto()
    ROI_CIRCLE = auto()
    ROI_POLYGON = auto()
    ROI_SEGMENTED_LINE = auto()
    ROI_LINE = auto()

    @classmethod
    def from_roi(cls, roi: QRoi) -> MouseMode:
        if isinstance(roi, QRectangleRoi):
            return cls.ROI_RECTANGLE
        if isinstance(roi, QRotatedRectangleRoi):
            return cls.ROI_ROTATED_RECTANGLE
        if isinstance(roi, QEllipseRoi):
            return cls.ROI_ELLIPSE
        if isinstance(roi, QRotatedEllipseRoi):
            return cls.ROI_ROTATED_ELLIPSE
        if isinstance(roi, QPointRoi):
            return cls.ROI_POINT
        if isinstance(roi, QPointsRoi):
            return cls.ROI_POINTS
        if isinstance(roi, QCircleRoi):
            return cls.ROI_CIRCLE
        if isinstance(roi, QPolygonRoi):
            return cls.ROI_POLYGON
        if isinstance(roi, QSegmentedLineRoi):
            return cls.ROI_SEGMENTED_LINE
        if isinstance(roi, QLineRoi):
            return cls.ROI_LINE
        raise ValueError(f"Unknown ROI type: {type(roi)}")


SIMPLE_ROI_MODES = frozenset({
    MouseMode.ROI_RECTANGLE,
    MouseMode.ROI_ROTATED_RECTANGLE,
    MouseMode.ROI_ELLIPSE,
    MouseMode.ROI_ROTATED_ELLIPSE,
    MouseMode.ROI_CIRCLE,
    MouseMode.ROI_POINT,
    MouseMode.ROI_LINE
})  # fmt: skip
MULTIPOINT_ROI_MODES = frozenset({
    MouseMode.ROI_POINTS, MouseMode.ROI_POLYGON, MouseMode.ROI_SEGMENTED_LINE
})  # fmt: skip
ROI_MODES = SIMPLE_ROI_MODES | MULTIPOINT_ROI_MODES
MULTIPOINT_ROI_CLASSES = (QPolygonRoi, QSegmentedLineRoi, QPointsRoi)



import math
from himena.standards import roi
import pytest

def test_span_roi():
    r0 = roi.SpanRoi(start=0, end=1)
    r1 = r0.shifted(2)
    assert r1.start == 2
    assert r1.end == 3
    assert r0.width() == 1
    assert r1.width() == 1

def test_point_roi_1d():
    r0 = roi.PointRoi1D(x=0.6)
    r1 = r0.shifted(2)
    assert r1.x == pytest.approx(2.6)

def test_points_roi_1d():
    r0 = roi.PointsRoi1D(xs=[0.6, 1.2])
    r1 = r0.shifted(2)
    assert list(r1.xs) == pytest.approx([2.6, 3.2])

def test_rectangle():
    r0 = roi.RectangleRoi(x=0, y=0, width=1, height=2)
    r1 = r0.shifted(2, 3)
    assert r1.x == 2
    assert r1.y == 3
    assert r1.width == 1
    assert r1.height == 2
    assert r1.area() == 2
    assert r1.to_mask((5, 5)).sum() == 2

def test_rotated_rectangle():
    r0 = roi.RotatedRectangleRoi(start=(1, 1), end=(2, 2), width=1)
    r1 = r0.shifted(2, 3)
    assert r0.length() == pytest.approx(math.sqrt(2), rel=1e-5)
    assert r1.start == (3, 4)
    assert r1.end == (4, 5)
    assert r0.angle() == pytest.approx(-45)
    assert r1.angle_radian() == pytest.approx(-math.pi / 4, rel=1e-5)
    assert r0.to_mask((5, 5)).sum() < 3

def test_ellipse():
    r0 = roi.EllipseRoi(x=0, y=0, width=6, height=8)
    r1 = r0.shifted(2, 3)
    assert r1.x == 2
    assert r1.y == 3
    assert r1.width == 6
    assert r1.height == 8
    assert r0.center() == pytest.approx((3, 4), rel=1e-5)
    assert r1.area() == pytest.approx(math.pi * 3 * 4, rel=1e-5)
    assert r1.to_mask((20, 20)).sum() <= math.pi * 3 * 4
    assert r0.eccentricity() == pytest.approx(math.sqrt(7 / 16), rel=1e-5)
    assert r0.to_mask((20, 20)).sum() <= math.pi * 3 * 4
    assert r0.to_mask((20, 20)).sum() == r1.to_mask((20, 20)).sum()

    # flat ellipse
    rflat = roi.EllipseRoi(x=0, y=4, width=6, height=0)
    assert rflat.area() == 0
    assert rflat.to_mask((20, 20)).sum() == 0
    assert rflat.eccentricity() == 0.0

    rflat = roi.EllipseRoi(x=0, y=4, width=0, height=4)
    assert rflat.area() == 0
    assert rflat.to_mask((20, 20)).sum() == 0
    assert rflat.eccentricity() == 0.0

    rflat = roi.EllipseRoi(x=0, y=4, width=0, height=0)
    assert rflat.area() == 0
    assert rflat.to_mask((20, 20)).sum() == 0
    assert rflat.eccentricity() == 1.0

def test_rotated_ellipse():
    r0 = roi.RotatedEllipseRoi(start=(1, 1), end=(4, 4), width=2)
    r1 = r0.shifted(2, 3)
    assert r0.length() == pytest.approx(3 * math.sqrt(2), rel=1e-5)
    assert r0.area() == pytest.approx(r1.area(), rel=1e-5)
    assert r0.area() == pytest.approx(math.pi * 3 * math.sqrt(2) / 2, rel=1e-5)
    assert r0.eccentricity() == pytest.approx(math.sqrt(7 / 9), rel=1e-5)

def test_circle():
    r0 = roi.CircleRoi(x=0, y=0, radius=5)
    r1 = r0.shifted(10, 8)
    assert r1.x == 10
    assert r1.y == 8
    assert r1.radius == 5
    assert r0.area() == pytest.approx(math.pi * 25, rel=1e-5)
    assert r0.to_mask((20, 20)).sum() < math.pi * 25

def test_point_roi_2d():
    r0 = roi.PointRoi2D(x=0.6, y=1.2)
    r1 = r0.shifted(2, 3)
    assert r1.x == pytest.approx(2.6)
    assert r1.y == pytest.approx(4.2)
    assert r0.bbox().width == 0
    assert r0.to_mask((5, 5)).sum() == 1

def test_points_roi_2d():
    r0 = roi.PointsRoi2D(xs=[0.6, 1.2], ys=[1.2, 2.4])
    r1 = r0.shifted(2, 3)
    assert list(r1.xs) == pytest.approx([2.6, 3.2])
    assert list(r1.ys) == pytest.approx([4.2, 5.4])
    assert r0.bbox().width == pytest.approx(0.6, rel=1e-5)
    assert r0.bbox().height == pytest.approx(1.2, rel=1e-5)
    assert r0.to_mask((5, 5)).sum() == 2



from __future__ import annotations

import math
from typing import TYPE_CHECKING
from qtpy import QtWidgets as QtW, QtGui, QtCore
from himena.consts import DefaultFontFamily
from himena.utils.enum import StrEnum

if TYPE_CHECKING:
    from ._graphics_view import QImageGraphicsView


class QScaleBarItem(QtW.QGraphicsItem):
    """A scale bar item for a QGraphicsView."""

    def __init__(self, view: QImageGraphicsView):
        super().__init__()
        self._view = view
        self._scale = 1.0
        self._unit = "px"
        self._point_size = 10
        self._font = QtGui.QFont(DefaultFontFamily)
        self._color = QtGui.QColor(255, 255, 255)
        self._anchor_offset_px = QtCore.QPointF(8, 8)
        self._bar_size_px = QtCore.QPointF(20, 3)
        self._bar_rect = QtCore.QRectF(0, 0, 0, 0)
        self._bounding_rect = QtCore.QRectF(0, 0, 0, 0)
        self._text_visible = True
        self._auto_adjust_size = True
        self._current_adjust_factor = 1.0
        self._anchor = ScaleBarAnchor.BOTTOM_RIGHT
        self._scale_bar_type = ScaleBarType.SHADOWED

    def update_rect(self, qrect: QtCore.QRectF):
        vw, vh = self._view.width(), self._view.height()
        scale = self._view.transform().m11()
        self._font.setPointSizeF(self._point_size)
        text_height = self._point_size / scale * 1.5 if self._text_visible else 0.0
        box_height = self._bar_size_px.y() / scale
        off = self._anchor_offset_px / scale

        if self._auto_adjust_size:
            _adjust_factor = _find_nice_factor(scale, self._scale)
            box_width = _adjust_factor / self._scale
            self._current_adjust_factor = _adjust_factor
        else:
            box_width = self._bar_size_px.x()  # do not divide by `scale`

        if self._anchor is ScaleBarAnchor.TOP_LEFT:
            box_top_left = self._view.mapToScene(0, 0) + off
        elif self._anchor is ScaleBarAnchor.TOP_RIGHT:
            box_top_left = (
                self._view.mapToScene(vw, 0)
                - QtCore.QPointF(box_width, 0)
                + QtCore.QPointF(-off.x(), off.y())
            )
        elif self._anchor is ScaleBarAnchor.BOTTOM_LEFT:
            box_top_left = (
                self._view.mapToScene(0, vh)
                - QtCore.QPointF(0, box_height + text_height)
                + QtCore.QPointF(off.x(), -off.y())
            )
        else:
            box_top_left = (
                self._view.mapToScene(vw, vh)
                - QtCore.QPointF(box_width, box_height + text_height)
                - off
            )

        self._bar_rect = QtCore.QRectF(
            box_top_left.x(), box_top_left.y(), box_width, box_height
        )
        self.update()

    def scale_bar_text(self) -> str:
        """Text to display on the scale bar."""
        if self._auto_adjust_size:
            val = self._current_adjust_factor
            return f"{val} {self._unit}"
        return f"{int(round(self._bar_size_px.x() * self._scale))} {self._unit}"

    def bar_and_text_rect(self) -> tuple[QtCore.QRectF, QtCore.QRectF]:
        scale = self._view.transform().m11()
        text = self.scale_bar_text()
        metrics = QtGui.QFontMetricsF(self._font)
        width = (metrics.width(text) + 4) / scale
        height = metrics.height() / scale
        text_rect = QtCore.QRectF(
            self._bar_rect.center().x() - width / 2,
            self._bar_rect.bottom() + 1 / scale,
            width,
            height,
        )
        bar_rect = QtCore.QRectF(self._bar_rect)
        if width > bar_rect.width():
            dw = (width - bar_rect.width()) / 2
            bar_rect.adjust(-dw, 0, dw, 0)
        return bar_rect, text_rect

    def paint(self, painter, option, widget=None):
        painter.setFont(self._font)
        text = self.scale_bar_text()
        if self._scale_bar_type is ScaleBarType.SHADOWED:
            self.draw_shadowed_scale_bar(painter, text)
        elif self._scale_bar_type is ScaleBarType.BACKGROUND:
            self.draw_backgrounded_scale_bar(painter, text)
        else:
            self.draw_simple_scale_bar(painter, text)

    def update_scale_bar(
        self,
        scale: float | None = None,
        unit: str | None = None,
        color: QtGui.QColor | None = None,
        anchor: ScaleBarAnchor | None = None,
        type: ScaleBarType | None = None,
        visible: bool | None = None,
        text_visible: bool | None = None,
    ):
        if scale is not None:
            if scale <= 0:
                raise ValueError("Scale must be positive")
            self._scale = scale
        if unit is not None:
            self._unit = unit
        if anchor is not None:
            self._anchor = ScaleBarAnchor(anchor)
        if type is not None:
            self._scale_bar_type = ScaleBarType(type)
        if color is not None:
            self._color = QtGui.QColor(color)
        if visible is not None:
            self.setVisible(visible)
        if text_visible is not None:
            self._text_visible = text_visible
        self.update_rect(self._view.sceneRect())

    def boundingRect(self):
        return self._bounding_rect

    def set_bounding_rect(self, rect: QtCore.QRectF):
        self._bounding_rect = QtCore.QRectF(rect)
        self.update()

    def draw_simple_scale_bar(self, painter: QtGui.QPainter, text: str):
        painter.setPen(QtGui.QPen(self._color, 0))
        bar_rect, text_rect = self.bar_and_text_rect()
        painter.setBrush(self._color)
        painter.drawRect(bar_rect)
        if self._text_visible:
            self._draw_text_original_scale(painter, text_rect, text)

    def draw_shadowed_scale_bar(self, painter: QtGui.QPainter, text: str):
        bar_rect, text_rect = self.bar_and_text_rect()
        shadow_rect = bar_rect.translated(0, bar_rect.height())
        color_shadow = (
            QtGui.QColor(0, 0, 0)
            if self._color.lightness() > 128
            else QtGui.QColor(255, 255, 255)
        )
        painter.setPen(QtGui.QPen(color_shadow, 0))
        painter.setBrush(color_shadow)
        painter.drawRect(shadow_rect)
        _2 = 2 / self._view.transform().m11()
        if self._text_visible:
            self._draw_text_original_scale(painter, text_rect.translated(0, _2), text)
        self.draw_simple_scale_bar(painter, text)
        painter.setPen(QtGui.QPen(self._color, 0))
        painter.setBrush(self._color)
        painter.drawRect(bar_rect)
        if self._text_visible:
            self._draw_text_original_scale(painter, text_rect, text)

    def draw_backgrounded_scale_bar(self, painter: QtGui.QPainter, text: str):
        bar_rect, text_rect = self.bar_and_text_rect()
        color_bg = (
            QtGui.QColor(0, 0, 0)
            if self._color.lightness() > 128
            else QtGui.QColor(255, 255, 255)
        )
        painter.setPen(QtGui.QPen(color_bg, 0))
        painter.setBrush(color_bg)
        _4 = 1 / self._view.transform().m11() * 4
        rect_bg = bar_rect.united(text_rect).adjusted(-_4, -_4, _4, _4)
        painter.drawRect(rect_bg)
        painter.setPen(QtGui.QPen(self._color, 0))
        painter.setBrush(self._color)
        if self._text_visible:
            self._draw_text_original_scale(painter, text_rect, text)
        self.draw_simple_scale_bar(painter, text)

    def _draw_text_original_scale(
        self,
        painter: QtGui.QPainter,
        rect: QtCore.QRectF,
        text: str,
    ):
        tr = painter.transform()
        painter.resetTransform()
        top_left = self._view.mapFromScene(rect.topLeft())
        bottom_right = self._view.mapFromScene(rect.bottomRight())
        rect = QtCore.QRectF(QtCore.QPointF(top_left), QtCore.QPointF(bottom_right))
        painter.drawText(rect, text)
        painter.setTransform(tr)


class ScaleBarType(StrEnum):
    SIMPLE = "simple"
    SHADOWED = "shadowed"
    BACKGROUND = "background"


class ScaleBarAnchor(StrEnum):
    TOP_LEFT = "top_left"
    TOP_RIGHT = "top_right"
    BOTTOM_LEFT = "bottom_left"
    BOTTOM_RIGHT = "bottom_right"


def _find_nice_factor(zoom_factor: float, image_scale: float) -> float:
    """Find a nice width for the scale bar."""
    # physical_width should be 1, 2, 5 x 10^N
    _n_pixel_for_1_unit = 1 / image_scale
    _min_logical_size = 40
    _max_logical_size = 100
    _screen_size = _n_pixel_for_1_unit * zoom_factor
    # find the smallest 1, 2, 5 x 10^N that is larger than _logical_size
    _n = -math.floor(math.log10(_screen_size)) + 1
    # now, _logical_size * 10^_n is in the range [10, 100)
    _multiplied = _screen_size * 10**_n
    _factor_int = 1
    if _multiplied < _min_logical_size / 2:
        _factor_int = 5
    elif _multiplied < _min_logical_size:
        _factor_int = 2
    elif _multiplied >= _max_logical_size * 2:
        _factor_int = 2
        _n -= 1
    elif _multiplied >= _max_logical_size:
        _factor_int = 5
        _n -= 1
    else:
        _factor_int = 1
    return _factor_int * 10**_n



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, Iterator
import uuid
from himena.utils.misc import is_subtype
from himena.workflow._base import WorkflowStep
from himena.workflow._reader import ReaderMethod
from himena.workflow._command import CommandExecution

if TYPE_CHECKING:
    from typing import Self
    from himena.widgets import MainWindow


@dataclass
class ActionMatcher(ABC):
    """Class that determine if a workflow step satisfies a specific condition."""

    model_type: str
    """The model type of the output window."""

    @abstractmethod
    def match(self, model_type: str, step: WorkflowStep) -> bool:
        """Check if the action matcher matches the given workflow step."""


@dataclass
class ReaderMatcher(ActionMatcher):
    """Matches workflow steps that are data loaders."""

    plugins: list[str] = field(default_factory=list)

    def match(self, model_type: str, step: WorkflowStep) -> bool:
        if is_subtype(model_type, self.model_type) and isinstance(step, ReaderMethod):
            return (
                step.plugin is None
                or len(self.plugins) == 0
                or step.plugin in self.plugins
            )
        return False


@dataclass
class CommandMatcher(ActionMatcher):
    """Matches workflow steps that are command-line based."""

    command_id: str

    def match(self, model_type: str, step: WorkflowStep) -> bool:
        if (
            is_subtype(model_type, self.model_type)
            and isinstance(step, CommandExecution)
            and step.command_id == self.command_id
        ):
            # TODO: match parameters and contexts
            return True
        return False


@dataclass
class Suggestion(ABC):
    @abstractmethod
    def execute(self, main_window: MainWindow, step: WorkflowStep) -> None:
        """Execute the suggestion in the given main window."""

    @abstractmethod
    def get_title(self, main_window: MainWindow) -> str:
        """Get the title of the suggestion."""

    @abstractmethod
    def get_tooltip(self, main_window: MainWindow) -> str:
        """Get the tooltip for the suggestion."""

    def make_executor(
        self, main_window: MainWindow, step: WorkflowStep
    ) -> Callable[[], None]:
        """Create an executor for the suggestion."""
        return lambda: self.execute(main_window, step)


@dataclass
class CommandSuggestion(Suggestion):
    """Next action that is a command execution."""

    command_id: str
    """The ID of the command to execute."""
    window_id: Callable[[WorkflowStep], uuid.UUID] | None = None
    """The model context to use for the command execution, if any."""
    defaults: dict[str, Callable[[WorkflowStep], Any]] = field(default_factory=dict)
    """Default parameter overrides."""

    def execute(self, main_window: MainWindow, step: WorkflowStep) -> None:
        """Execute the suggestion in the given main window."""
        params = {}
        for key, factory in self.defaults.items():
            value = factory(step)
            params[key] = value
        if self.window_id is not None:
            window_context = main_window.window_for_id(self.window_id(step))
        else:
            window_context = None
        main_window.exec_action(
            self.command_id, window_context=window_context, with_defaults=params
        )

    def get_title(self, main_window: MainWindow) -> str:
        """Get the title of the command suggestion."""
        if cmd := main_window.model_app.registered_actions.get(self.command_id, None):
            return cmd.title
        return "Unknown Command"

    def get_tooltip(self, main_window: MainWindow) -> str:
        """Get the tooltip for the command suggestion."""
        if cmd := main_window.model_app.registered_actions.get(self.command_id, None):
            return cmd.tooltip
        return "No tooltip available for this command."


@dataclass
class ActionHint:
    matcher: ActionMatcher
    """The matcher that determines if this action hint is applicable."""
    suggestion: Suggestion
    """The next action to take if this action hint is applicable."""


class ActionHintRegistry:
    def __init__(self):
        self._rough_map: dict[str, list[ActionHint]] = {}

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}:" + "".join(
            f"\n- {each!r}" for each in self.iter_all()
        )

    def iter_all(self) -> Iterator[ActionHint]:
        """Iterate over all action hints in the registry."""
        for hints in self._rough_map.values():
            yield from hints

    def add_hint(self, matcher: ActionMatcher, suggestion: Suggestion) -> None:
        """Add a matcher to the registry."""
        ancestor_type = matcher.model_type.split(".")[0]
        if ancestor_type not in self._rough_map:
            self._rough_map[ancestor_type] = []
        self._rough_map[ancestor_type].append(ActionHint(matcher, suggestion))

    def when_command_executed(
        self,
        model_type: str,
        command_id: str,
    ) -> ActionMatcherInterface:
        """Create an interface for adding command suggestions.

        Examples
        --------
        >>> (
        ...     reg.when_command_executed("table", "sort-table")
        ...        .add_command_suggestion("scatter-plot")
        ...        .add_command_suggestion("line-plot")
        ... )
        """
        matcher = CommandMatcher(model_type=model_type, command_id=command_id)
        return ActionMatcherInterface(self, matcher)

    def when_reader_used(
        self,
        model_type: str,
        plugins: list[str] | None = None,
    ) -> ActionMatcherInterface:
        """Create an interface for adding reader suggestions.

        Examples
        --------
        >>> (
        ...     reg.when_reader_used("table")
        ...        .add_command_suggestion("load-csv")
        ... )
        """
        matcher = ReaderMatcher(model_type=model_type, plugins=plugins or [])
        return ActionMatcherInterface(self, matcher)

    def iter_suggestion(
        self, model_type: str, step: WorkflowStep
    ) -> Iterator[Suggestion]:
        """Get a list of matchers that match the given model type and step."""
        for ancestor_type, hints in self._rough_map.items():
            if model_type.startswith(ancestor_type):
                for hint in hints:
                    if hint.matcher.match(model_type, step):
                        yield hint.suggestion


class ActionMatcherInterface:
    def __init__(self, reg: ActionHintRegistry, matcher: ActionMatcher):
        self._registry = reg
        self._matcher = matcher

    def add_command_suggestion(
        self,
        command_id: str,
        window_id: Callable[[WorkflowStep], uuid.UUID] | None = None,
        defaults: dict[str, Callable[[WorkflowStep], Any]] | None = None,
    ) -> Self:
        """Add a suggestion to the registry."""
        suggestion = CommandSuggestion(
            command_id=command_id, window_id=window_id, defaults=defaults or {}
        )
        self._registry.add_hint(self._matcher, suggestion)
        return self



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtCore import Qt
from himena.consts import MonospaceFontFamily
from himena.plugins import AppActionRegistry
from himena.utils.entries import iter_plugin_info
from himena.qt.settings._shared import QInstruction

if TYPE_CHECKING:
    from himena.widgets import MainWindow


class QPluginListEditor(QtW.QWidget):
    """Widget to edit plugin list."""

    def __init__(self, ui: MainWindow):
        super().__init__()
        self._ui = ui
        self._app = ui.model_app
        layout = QtW.QVBoxLayout(self)
        _instruction1 = QInstruction(
            f"Select plugins to be included in <b><code>himena {self._app.name}</code>"
            "</b>.<br>Click <b>Apply</b> to save changes.",
        )
        self._plugins_editor = QAvaliablePluginsTree(self)
        self._plugins_editor.stateChanged.connect(self._enabled_apply_button)

        _instruction2 = QInstruction("List of '.py' files as unpackaged plugins.")
        self._additional_plugin_list = QtW.QPlainTextEdit()
        self._additional_plugin_list.setHorizontalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOff
        )
        self._additional_plugin_list.setWordWrapMode(QtGui.QTextOption.WrapMode.NoWrap)
        self._additional_plugin_list.setFont(QtGui.QFont(MonospaceFontFamily))
        self._additional_plugin_list.setFixedHeight(120)
        self._additional_plugin_list.setPlaceholderText("e.g. path/to/file.py")
        for plugin_name in ui.app_profile.plugins:
            if plugin_name.endswith(".py") and Path(plugin_name).exists():
                self._additional_plugin_list.appendPlainText(plugin_name + "\n")

        self._apply_button = QtW.QPushButton("Apply", self)
        self._button_group = QtW.QWidget(self)
        button_layout = QtW.QHBoxLayout(self._button_group)
        button_layout.setContentsMargins(0, 0, 0, 0)
        button_layout.setAlignment(Qt.AlignmentFlag.AlignRight)
        button_layout.addWidget(self._apply_button)

        layout.addWidget(_instruction1)
        layout.addWidget(self._plugins_editor)
        layout.addWidget(_instruction2)
        layout.addWidget(self._additional_plugin_list)
        layout.addWidget(self._button_group)

        self._apply_button.clicked.connect(self._apply_changes)
        self._apply_button.setEnabled(False)
        self._additional_plugin_list.textChanged.connect(self._enabled_apply_button)

    def _apply_changes(self):
        plugins = self._plugins_editor.get_plugin_list()
        for line in self._additional_plugin_list.toPlainText().splitlines():
            line = line.strip()
            if line:
                plugins.append(line)
        new_prof = self._ui.app_profile.with_plugins(plugins)
        AppActionRegistry.instance().install_to(self._app)
        new_prof.save()
        self._apply_button.setEnabled(False)

    def _enabled_apply_button(self):
        self._apply_button.setEnabled(True)


class QAvaliablePluginsTree(QtW.QTreeWidget):
    stateChanged = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        font = QtGui.QFont(MonospaceFontFamily)
        self.setFont(font)
        self.setColumnCount(1)
        self.setHeaderHidden(True)
        self.setIndentation(10)
        last_distribution: str | None = None
        current_toplevel_item: QtW.QTreeWidgetItem | None = None

        reg = AppActionRegistry.instance()
        installed_plugins = reg.installed_plugins
        _flags = Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsEnabled

        for info in iter_plugin_info():
            if info.distribution != last_distribution:
                last_distribution = info.distribution
                if current_toplevel_item is not None:
                    current_toplevel_item.setExpanded(True)
                current_toplevel_item = QtW.QTreeWidgetItem([info.distribution])
                current_toplevel_item.setFlags(_flags)
                current_toplevel_item.setCheckState(0, Qt.CheckState.Checked)
                self.addTopLevelItem(current_toplevel_item)
            item = QtW.QTreeWidgetItem([f"{info.name} ({info.place})", info.place])
            item.setFlags(_flags)
            if info.place in installed_plugins:
                item.setCheckState(0, Qt.CheckState.Checked)
            else:
                item.setCheckState(0, Qt.CheckState.Unchecked)
                current_toplevel_item.setCheckState(0, Qt.CheckState.Unchecked)
            current_toplevel_item.addChild(item)
        if current_toplevel_item is not None:
            current_toplevel_item.setExpanded(True)
        self.itemChanged.connect(self._on_item_changed)

    def get_plugin_list(self) -> list[str]:
        """Return the list of plugin IDs."""
        plugins: list[str] = []
        for i in range(self.topLevelItemCount()):
            dist_item = self.topLevelItem(i)
            for j in range(dist_item.childCount()):
                plugin_item = dist_item.child(j)
                if plugin_item.checkState(0) == Qt.CheckState.Checked:
                    plugins.append(plugin_item.text(1))
        return plugins

    def _on_item_changed(self, item: QtW.QTreeWidgetItem, column: int):
        if item.parent() is None:
            for i in range(item.childCount()):
                item.child(i).setCheckState(0, item.checkState(0))
        self.stateChanged.emit()



from __future__ import annotations

import math
import numpy as np
from numpy.typing import NDArray


def polygon_mask(
    shape: tuple[int, int], vertices: NDArray[np.number]
) -> NDArray[np.bool_]:
    """Create a binary mask of a polygon.

    Parameters
    ----------
    shape : tuple[int, int]
        Shape of the mask.
    vertices : np.ndarray
        Nx2 array of the vertices of the polygon.

    Returns
    -------
    mask : np.ndarray
        Binary mask of the polygon.
    """
    x, y = np.indices(shape)
    points = np.vstack((x.ravel(), y.ravel())).T
    mask = points_in_poly(points, vertices).reshape(shape)
    return mask


# This function is copied from https://github.com/napari/napari/blob/main/napari/layers/shapes/_shapes_utils.py
def points_in_poly(points: NDArray[np.number], vertices: NDArray[np.number]):
    """Tests points for being inside a polygon using the ray casting algorithm

    Parameters
    ----------
    points : np.ndarray
        Mx2 array of points to be tested
    vertices : np.ndarray
        Nx2 array of the vertices of the polygon.

    Returns
    -------
    inside : np.ndarray
        Length M boolean array with `True` for points inside the polygon
    """
    n_verts = len(vertices)
    inside = np.zeros(len(points), dtype=bool)
    j = n_verts - 1
    for i in range(n_verts):
        # Determine if a horizontal ray emanating from the point crosses the
        # line defined by vertices i-1 and vertices i.
        cond_1 = np.logical_and(
            vertices[i, 1] <= points[:, 1], points[:, 1] < vertices[j, 1]
        )
        cond_2 = np.logical_and(
            vertices[j, 1] <= points[:, 1], points[:, 1] < vertices[i, 1]
        )
        cond_3 = np.logical_or(cond_1, cond_2)
        d = vertices[j] - vertices[i]
        # Prevents floating point imprecision from generating false positives
        tolerance = 1e-12
        d = np.where(abs(d) < tolerance, 0, d)
        if d[1] == 0:
            # If y vertices are aligned avoid division by zero
            cond_4 = d[0] * (points[:, 1] - vertices[i, 1]) > 0
        else:
            cond_4 = points[:, 0] < (
                d[0] * (points[:, 1] - vertices[i, 1]) / d[1] + vertices[i, 0]
            )
        cond_5 = np.logical_and(cond_3, cond_4)
        inside[cond_5] = 1 - inside[cond_5]
        j = i

    # If the number of crossings is even then the point is outside the polygon,
    # if the number of crossings is odd then the point is inside the polygon

    return inside


def eccentricity(a: float, b: float) -> float:
    a, b = sorted([a, b])
    if a == 0:
        if b == 0:
            return 1.0
        return 0.0
    if b == 0:
        return 0.0
    return math.sqrt(1 - a**2 / b**2)



from himena import StandardType
from himena.plugins import register_widget_class
from himena_builtins.qt.widgets import _image_commands
from himena_builtins.qt.widgets._image_components import QImageGraphicsView

from himena_builtins.qt.widgets.image import (
    QImageView,
    QImageLabelView,
    ImageViewConfigs,
)
from himena_builtins.qt.widgets._image_components import _roi_items as QtRois

register_widget_class(
    StandardType.IMAGE, QImageView, priority=50, plugin_configs=ImageViewConfigs()
)
register_widget_class(StandardType.IMAGE_LABELS, QImageLabelView, priority=50)

del _image_commands

__all__ = [
    "QImageGraphicsView",
    "QImageView",
    "QImageLabelView",
    "ImageViewConfigs",
    "QtRois",
]



from typing import Sequence
from cmap import Colormap
import numpy as np
from numpy.testing import assert_array_equal
import pytest
from pytestqt.qtbot import QtBot
from qtpy.QtCore import Qt
from qtpy import QtWidgets as QtW
from himena import MainWindow, StandardType
from himena.standards.model_meta import ArrayMeta, DimAxis, ImageChannel, ImageMeta
from himena.testing import WidgetTester
from himena_builtins.qt.array import QArrayView
from himena_builtins.tools.array import _broadcast_arrays

_Ctrl = Qt.KeyboardModifier.ControlModifier

def test_array_view(himena_ui: MainWindow, qtbot: QtBot):
    with WidgetTester(QArrayView(himena_ui)) as tester:
        qtbot.addWidget(tester.widget)
        tester.widget.show()
        table = tester.widget._table
        table.update()
        table.model().data(table.model().index(0, 0), Qt.ItemDataRole.ToolTipRole)
        table.model().headerData(0, Qt.Orientation.Horizontal, Qt.ItemDataRole.ToolTipRole)
        table.model().headerData(0, Qt.Orientation.Vertical, Qt.ItemDataRole.ToolTipRole)
        table.selection_model.set_ranges([(slice(1, 2), slice(1, 3))])
        table._copy_data()
        table._make_context_menu()
        tester.update_model(value=np.arange(72).reshape(3, 2, 3, 4))
        assert len(tester.widget._spinboxes) == 2
        assert tester.widget._spinboxes[0].maximum() == 2
        assert tester.widget._spinboxes[1].maximum() == 1
        tester.widget._spinboxes[0].setValue(1)
        tester.widget._spinboxes[0].setValue(2)
        tester.widget._spinboxes[1].setValue(1)
        tester.widget._spinboxes[1].setValue(0)
        tester.widget.selection_model.set_ranges([(slice(1, 2), slice(1, 3))])
        qtbot.keyClick(tester.widget, Qt.Key.Key_C, _Ctrl)
        old, new = tester.cycle_model()
        assert np.all(old.value == new.value)
        assert new.metadata.selections == [((1, 2), (1, 3))]
        assert old.metadata.selections == new.metadata.selections

def test_structured(himena_ui: MainWindow, qtbot: QtBot):
    with WidgetTester(QArrayView(himena_ui)) as tester:
        qtbot.addWidget(tester.widget)
        tester.update_model(
            value=np.array(
                [("a", 1, 2.3, True), ("b", -2, -0.04, False)],
                dtype=[("name", "U1"), ("value", "i4"), ("float", "f4"), ("b", "bool")]
            )
        )
        assert tester.to_model().value.shape == (2,)
        assert tester.to_model().value.dtype.names == ("name", "value", "float", "b")

        tester.widget.selection_model.set_ranges([(slice(1, 2), slice(1, 3))])
        qtbot.keyClick(tester.widget, Qt.Key.Key_C, _Ctrl)
        old, new = tester.cycle_model()
        assert np.all(old.value == new.value)
        assert new.metadata.selections == [((1, 2), (1, 3))]
        assert old.metadata.selections == new.metadata.selections

def test_copy_and_paste(himena_ui: MainWindow, qtbot: QtBot):
    with WidgetTester(QArrayView(himena_ui)) as tester:
        qtbot.addWidget(tester.widget)
        tester.update_model(value=np.zeros((3, 5, 3, 4), dtype=np.int32))
        tester.widget.set_indices(1, 2)
        tester.widget.selection_model.current_index = (2, 0)
        qtbot.keyClick(tester.widget._table, Qt.Key.Key_F2)
        tester.widget._table.set_string_input(2, 0, "4")
        assert tester.to_model().value[1, 2, 2, 0] == 4
        # test copy and paste
        tester.widget.selection_model.set_ranges([(slice(2, 3), slice(0, 1))])
        tester.widget._table._copy_data()
        QtW.QApplication.processEvents()
        assert QtW.QApplication.clipboard().text() == "4"
        tester.widget.selection_model.set_ranges([(slice(0, 2), slice(1, 3))])
        tester.widget._table._paste_from_clipboard()
        assert_array_equal(tester.to_model().value[1, 2, 0:2, 1:3], 4)
        QtW.QApplication.clipboard().setText("1\t2\n3\t4")
        QtW.QApplication.processEvents()
        tester.widget.selection_model.set_ranges([(slice(0, 1), slice(0, 1))])
        tester.widget._table._paste_from_clipboard()
        assert_array_equal(tester.to_model().value[1, 2, 0:2, 0:2], [[1, 2], [3, 4]])

def test_binary_operations(himena_ui: MainWindow):
    win = himena_ui.add_object(np.arange(24, dtype=np.uint16).reshape(2, 3, 4), type=StandardType.ARRAY)
    model = win.to_model()
    himena_ui.exec_action(
        "builtins:array:binary-operation",
        with_params={"x": model, "y": model, "operation": "sub", "result_dtype": "input"},
    )
    himena_ui.exec_action(
        "builtins:array:binary-operation",
        with_params={"x": model, "y": model, "operation": "sub", "result_dtype": "float32"}
    )
    himena_ui.exec_action(
        "builtins:array:binary-operation",
        with_params={"x": model, "y": model, "operation": "sub", "result_dtype": "float64"}
    )


def test_array_commands(himena_ui: MainWindow):
    win = himena_ui.add_object(np.arange(24).reshape(2, 3, 4), type=StandardType.ARRAY)
    himena_ui.exec_action("builtins:array:duplicate-slice")
    assert himena_ui.current_model.value.shape == (3, 4)
    assert_array_equal(himena_ui.current_model.value, np.arange(12).reshape(3, 4))

    himena_ui.add_object(
        np.arange(24).reshape(2, 3, 4),
        type=StandardType.IMAGE,
        metadata=ImageMeta(
            channel_axis=0,
            channels=[ImageChannel(colormap="red"), ImageChannel(colormap="green")],
        )
    )
    himena_ui.exec_action("builtins:array:duplicate-slice")
    assert himena_ui.current_model.value.shape == (3, 4)
    assert himena_ui.current_model.metadata.channel_axis is None
    assert himena_ui.current_model.metadata.channels[0].colormap == Colormap("red")
    assert_array_equal(himena_ui.current_model.value, np.arange(12).reshape(3, 4))

    himena_ui.current_window = win
    win.update_model(
        win.to_model().with_metadata(
            ArrayMeta(
                axes=_make_axes("tyx"),
                selections=[((1, 2), (1, 3))],
            )
        )
    )
    himena_ui.exec_action("builtins:array:crop")
    assert himena_ui.current_model.value.shape == (2, 1, 2)
    himena_ui.current_window = win
    himena_ui.exec_action(
        "builtins:array:crop-nd", with_params={
            "axis_0": (0, 1),
            "axis_1": (1, 2),
            "axis_2": (0, 2),
        }
    )
    assert himena_ui.current_model.value.shape == (1, 1, 2)

    himena_ui.exec_action("builtins:array:astype", with_params={"dtype": "float32"})
    himena_ui.current_window = win
    himena_ui.exec_action(
        "builtins:array:set-scale",
        with_params={"axis_2": "1.4", "axis_1": "1.0 um", "axis_0": "0.5um"}
    )
    meta_new = win.to_model().metadata
    assert isinstance(meta_new, ArrayMeta)
    meta_new.axes[0].scale == pytest.approx(0.5)
    meta_new.axes[0].unit == "um"
    meta_new.axes[1].scale == pytest.approx(1.0)
    meta_new.axes[1].unit == "um"
    meta_new.axes[2].scale == pytest.approx(1.4)
    meta_new.axes[2].unit == ""
    himena_ui.exec_action("builtins:array:with-axes", with_params={"axis_0": "t", "axis_1": "y", "axis_2": "x"})
    meta_new = win.to_model().metadata
    assert isinstance(meta_new, ArrayMeta)
    assert meta_new.axes[0].name == "t"
    assert meta_new.axes[1].name == "y"
    assert meta_new.axes[2].name == "x"

def test_copy_on_write(himena_ui: MainWindow, qtbot: QtBot):
    view = QArrayView(himena_ui)
    qtbot.addWidget(view)
    with WidgetTester(view) as tester:
        tester.update_model(value=np.zeros((3, 4)), type="array")

        array_orig = np.array([[1, 2], [3, 4]])
        tester.update_model(value=array_orig, type=StandardType.ARRAY)
        view.array_update((slice(0, 1), slice(1, 2)), 10)
        assert_array_equal(view.to_model().value, [[1, 10], [3, 4]])
        assert_array_equal(array_orig, [[1, 2], [3, 4]])

        view_other = QArrayView(himena_ui)
        view_other.update_model(view.to_model())
        view.array_update((slice(0, 1), slice(0, 1)), 20)
        assert_array_equal(view.to_model().value, [[20, 10], [3, 4]])
        assert_array_equal(view_other.to_model().value, [[1, 10], [3, 4]])

@pytest.mark.parametrize(
    "shape0, shape1, axes0, axes1, expected_shape, expected_axes",
    [
        ((2, 3), (2, 3), "yx", "yx", (2, 3), "yx"),
        ((2, 3), (3,), "yx", "x", (2, 3), "yx"),
        ((2,), (2, 3), "y", "yx", (2, 3), "yx"),
        ((2, 5, 4, 3), (2, 4, 3), "tzyx", "tyx", (2, 5, 4, 3), "tzyx"),
        ((5,), (2, 5), "t", None, (2, 5), None),
    ]
)
def test_broadcast_array(
    shape0: tuple[int, ...],
    shape1: tuple[int, ...],
    axes0: Sequence[str],
    axes1: Sequence[str],
    expected_shape: tuple[int, ...],
    expected_axes: Sequence[str],
):
    a_out, b_out, axes = _broadcast_arrays(
        np.zeros(shape0),
        np.zeros(shape1),
        _make_axes(axes0),
        _make_axes(axes1),
    )
    out = a_out + b_out
    assert out.shape == expected_shape
    if expected_axes is None:
        assert axes is None
    else:
        assert [a.name for a in axes] == list(expected_axes)

def _make_axes(names: Sequence[str] | None) -> list[DimAxis] | None:
    if names is None:
        return None
    return [DimAxis(name=name) for name in names]
